from __future__ import annotations

import abc
from collections import namedtuple
from typing import TYPE_CHECKING

import e3.log
from e3.error import E3Error

if TYPE_CHECKING:
    from typing import Any, Optional
    from e3.store.cache.backends.base import Cache

logger = e3.log.getLogger("store")


class ResourceInfo(metaclass=abc.ABCMeta):
    """Object representing resource metadata.

    This is subclassed by all store drivers.
    """

    @abc.abstractproperty
    def uid(self) -> str:
        """Return an unique identifier.

        This is meant to be used to implement a cache system.
        """
        pass  # all: no cover

    @abc.abstractmethod
    def verify(self, resource_path: str) -> bool:
        """Verify that a downloaded resource is valid.

        This is meant to validate the resource fingerprint (e.g. sha1sum).
        :param resource_path: path to the downloaded resource
        """
        pass  # all: no cover


class CachedResource(namedtuple("CachedResource", ["uid", "local_path"])):
    """Cached information about an already downloaded resource."""

    __slots__ = ()


class StoreError(E3Error):
    pass


class Store(metaclass=abc.ABCMeta):
    def __init__(self, store_configuration: Any, cache_backend: Cache):
        """Initialize a Store object.

        :param store_configuration: configuration for this store
        :param cache_backend: a Cache instance
        :return:
        """
        self.store_configuration = store_configuration
        self.cache_backend = cache_backend

    @abc.abstractmethod
    def get_resource_metadata(self, query: dict) -> ResourceInfo:
        """Get a resource metadata from query.

        :param query: a dictionary containing store specific queries to
            identify a resource
        """
        pass  # all: no cover

    def download_resource(self, metadata: ResourceInfo, dest: str) -> Optional[str]:
        """Download a resource identified by its metadata in dest.

        :param metadata: resource metadata
        :param dest: directory where the resource will be stored
        :return: resource path
        """
        cached_data = self.cache_backend.get(metadata.uid)
        if cached_data is not None:
            try:
                local_path = cached_data.local_path
            except (TypeError, AttributeError):
                logger.warning("invalid cache entry for %s", metadata.uid)
            else:
                if not metadata.verify(local_path):
                    logger.warning("corrupted cache entry for %s", metadata.uid)
                    # Discard invalid cache entry
                    self.cache_backend.delete(metadata.uid)
                else:
                    return local_path

        # else not in cache
        local_path = self.download_resource_content(metadata, dest)
        if local_path is None:
            # cannot get the resource
            return None

        # store the resource in the cache
        self.cache_backend.set(
            metadata.uid, CachedResource(uid=metadata.uid, local_path=local_path)
        )
        return local_path

    @abc.abstractmethod
    def download_resource_content(
        self, metadata: ResourceInfo, dest: str
    ) -> Optional[str]:
        """Download a resource identified by its metadata in dest.

        The resource is supposed to be validated with metadata.verify
        once the download is completed.

        :param metadata: resource metadata
        :param dest: directory where the resource will be stored
        :return: resource path
        """
        pass  # all: no cover



from __future__ import annotations

import errno
import os
import sys
import tempfile
import time
from typing import TYPE_CHECKING

import e3.log
from e3.fs import mkdir, rm
from e3.store.cache.backends.base import DEFAULT_TIMEOUT, Cache

try:
    import pickle as pickle
except ImportError:  # defensive code
    import pickle

if TYPE_CHECKING:
    from typing import Any, IO


class FileCache(Cache):

    cache_suffix = ".cache"

    def __init__(self, cache_configuration: Any):
        super().__init__(cache_configuration)
        self.cache_dir = cache_configuration["cache_dir"]

    def clear(self) -> None:
        rm(os.path.join(self.cache_dir, "*.cache"))

    def delete(self, uid: str) -> None:
        rm(self.uid_to_file(uid))

    def _create_cache_dir(self) -> None:
        mkdir(self.cache_dir)

    def uid_to_file(self, uid: str) -> str:
        """Convert a resource uid to a cache file path.

        This backend assumes that the uid is a safe value for a file name.
        :param uid: the resource uid
        """
        return os.path.join(self.cache_dir, uid + self.cache_suffix)

    @staticmethod
    def _is_expired(fd: IO[bytes]) -> bool:
        """Determine if an open cache file has expired.

        Automatically delete the file if it has passed its expiry time.
        """
        exp = pickle.load(fd)
        if exp is not None and exp < time.time():
            fd.close()
            rm(fd.name)
            return True
        return False

    def get(self, uid: str, default: Any = None) -> Any:
        cache_file = self.uid_to_file(uid)
        try:
            with open(cache_file, "rb") as fd:
                if not self._is_expired(fd):
                    return pickle.load(fd)
        except OSError as err:
            if err.errno == errno.ENOENT:
                pass  # Cache file was removed after the exists check
        return default

    def set(self, uid: str, value: Any, timeout: int = DEFAULT_TIMEOUT) -> bool:
        # Make sure that the cache dir exists
        self._create_cache_dir()
        dest_file = self.uid_to_file(uid)

        tmp_file = tempfile.NamedTemporaryFile(dir=self.cache_dir, delete=False)
        try:
            tmp_file.write(
                pickle.dumps(self.get_expiry_time(timeout), pickle.HIGHEST_PROTOCOL)
            )
            tmp_file.write(pickle.dumps(value, pickle.HIGHEST_PROTOCOL))
        except Exception as err:
            tmp_file.close()
            e3.log.debug("error when setting %s in %s:\n%s", uid, dest_file, err)
            return False
        else:
            tmp_file.close()

            if sys.platform == "win32":  # unix: no cover
                # atomic rename does not work on windows if the dest file
                # already exist
                rm(dest_file)
            os.rename(tmp_file.name, dest_file)
            return True

        finally:
            rm(tmp_file.name)



from __future__ import annotations

import calendar
from datetime import datetime
import time
from typing import TYPE_CHECKING, overload

if TYPE_CHECKING:
    from typing import Optional


@overload
def timestamp_as_string(value: None) -> None:
    ...


@overload
def timestamp_as_string(value: float) -> str:
    ...


def timestamp_as_string(value: Optional[float]) -> Optional[str]:
    """Convert a timestamp into a human readable date/time.

    :param value: a timestamp or None
    :return: the string representing the timestamp or None if value is None
    """
    if value is None:
        return None
    return time.strftime("%Y-%m-%dT%H:%M:%S", time.gmtime(value))


def utc_timestamp() -> int:
    """Return the number of seconds since epoch UTC."""
    d = datetime.utcnow()
    return calendar.timegm(d.utctimetuple())



from __future__ import annotations

import ast
import os
import re
import sys
from enum import Enum
from dataclasses import asdict
from pprint import pformat
from typing import TYPE_CHECKING

import e3.log

if TYPE_CHECKING:
    from typing import List, Optional

logger = e3.log.getLogger("e3.sys")


class RewriteNodeError(Exception):
    pass


class RewriteImportRule:
    """Rewrite Import node from AST.

    Skip or reject names imported by::

        from <module> import <names>

    or directly::

        import <module>
    """

    class RuleAction(Enum):
        reject = 0
        skip = 1

    def __init__(
        self, module: str, name: str = ".*", action: Optional[RuleAction] = None
    ):
        """Initialize the object.

        :param module: regexp suitable for re.match() to match against the
            module name (note that the string is automatically surrounded by
            ^ and $)
        :param name: regexp suitable for re.match() to match against the names
            imported via "from <module> import <names>" (note that the
            string is automatically surrounded by ^ and $)
        :param action: skip (default) to avoid importing the name, or reject
            to raise a RewriteNodeError
        :raise: RewriteNodeError
        """
        self.module = module
        self.name = name
        self.action = action if action is not None else self.RuleAction.skip

    def rewrite_node(self, node: ast.stmt) -> ast.stmt:
        """Rewrite a node.

        :param node: ast node
        :return: a modified ast node
        """
        check_in_names = None

        if isinstance(node, ast.ImportFrom):

            # node: ImportFrom(module, names)
            # first check whether the module match our rule

            if TYPE_CHECKING:
                assert node.module is not None
            if re.match("^" + self.module + "$", node.module):

                # then we need to check whether our 'name' is in the
                # 'from' list (node.names)

                check_in_names = self.name

        elif isinstance(node, ast.Import):

            # node: Import(names)
            # We need to check whether our 'module' appears in the imported
            # module names

            check_in_names = self.module

        if check_in_names is not None:
            new_names = []
            for _, var in enumerate(node.names):  # type: ignore
                if re.match("^" + check_in_names + "$", var.name):
                    if self.action == self.RuleAction.skip:
                        # don't import this name
                        pass
                    elif self.action == self.RuleAction.reject:
                        raise RewriteNodeError(
                            "Rejected import found in ast: %s" % ast.dump(node)
                        )
                    else:  # defensive code
                        raise ValueError("unknown action %s", self.action)
                else:
                    new_names.append(var)
            node.names = new_names  # type: ignore
        return node


class RewriteImportNodeTransformer(ast.NodeTransformer):
    """Walk the AST applying a set of rules.

    Currently only the RewriteImportRule are supported.
    """

    def __init__(self, rules: List[RewriteImportRule]):
        """Load a set of rules.

        :param rules: list of rule objects
        """
        self.rules = rules

    def visit_ImportFrom(self, node: ast.stmt) -> ast.stmt:
        for rule in self.rules:
            node = rule.rewrite_node(node)
        return node

    def visit_Import(self, node: ast.stmt) -> ast.stmt:
        for rule in self.rules:
            node = rule.rewrite_node(node)
        return node


def version() -> str:
    import pkg_resources

    return pkg_resources.get_distribution("e3-core").version


def sanity_check() -> int:
    """Sanity check the E3 install."""
    errors = 0
    print("YAMLCheck:", end=" ")
    try:
        import yaml

        yaml.safe_dump({"Yaml": "works"})
        print("PASSED")
    except Exception:  # defensive code
        print("FAILED")
        errors += 1

    print("HashlibCheck:", end=" ")
    try:
        from e3.hash import sha1, md5

        sha1(__file__)
        md5(__file__)
        print("PASSED")
    except Exception:  # defensive code
        print("FAILED")
        errors += 1

    print("Version:", end=" ")
    try:
        print(version())
    except Exception:  # defensive code
        errors += 1
    return errors


def main() -> None:
    from e3.env import Env
    import e3.main

    m = e3.main.Main(platform_args=True)
    m.argument_parser.add_argument(
        "--platform-info",
        choices={"build", "host", "target"},
        help="Show build/host/target platform info",
    )
    m.argument_parser.add_argument(
        "--version", help="Show E3 version", action="store_true"
    )
    m.argument_parser.add_argument(
        "--check", help="Run e3 sanity checking", action="store_true"
    )
    m.argument_parser.add_argument(
        "--show-config", action="store_true", help="Show e3 config"
    )
    m.parse_args()

    if TYPE_CHECKING:
        assert m.args is not None

    if m.args.version:
        print(version())
        return
    elif m.args.check:
        errors = sanity_check()
        if errors:  # defensive code
            logger.error("sanity checking failed!")
            sys.exit(1)
        else:
            print("Everything OK!")
            return
    elif m.args.platform_info:
        print(getattr(Env(), m.args.platform_info))

    if m.args.show_config:
        print("[log]")
        for k, v in asdict(e3.log.log_config).items():
            print("{}: {}".format(k, pformat(v)))


def set_python_env(prefix: str) -> None:
    """Set environment for a Python distribution.

    :param prefix: root directory of the python distribution
    """
    import e3.env

    env = e3.env.Env()
    if sys.platform == "win32":  # unix: no cover
        env.add_path(prefix)
        env.add_path(os.path.join(prefix, "Scripts"))
    else:
        env.add_path(os.path.join(prefix, "bin"))
        env.add_dll_path(os.path.join(prefix, "lib"))


def interpreter(prefix: Optional[str] = None) -> str:
    """Return location of the Python interpreter.

    When there are both a python3 and python binary file return the path to
    the python3 binary.

    :param prefix: root directory of the python distribution. if None location
        of the current interpreter is returned
    :return: python executable path
    """
    if prefix is None:
        return sys.executable
    if sys.platform == "win32":  # unix: no cover
        python3 = os.path.join(prefix, "python3.exe")
        if os.path.exists(python3):
            return python3
        else:
            return os.path.join(prefix, "python.exe")
    else:  # windows: no cover
        python3 = os.path.join(prefix, "bin", "python3")
        if os.path.exists(python3):
            return python3
        else:
            return os.path.join(prefix, "bin", "python")


def python_script(name: str, prefix: Optional[str] = None) -> List[str]:
    """Return path to scripts contained in this Python distribution.

    :param name: the script name
    :param prefix: root directory of the Python distribution. if None the
        distribution currently used by this script will be used
    :return: a list that will be the prefix of your command line
    """

    def has_relative_python_shebang(file_script: str) -> bool:  # unix: no cover
        """Return True if the script contains #!python shebang.

        When producing relocatable python distribution we change the shebang
        to #!python. In that case prefix the command line with the current
        python interpreter.

        #!/path/to/python shebang should return false, as we don't need to
        return interpreter path.
        """
        with open(file_script, "rb") as f:
            content = f.read()
        return re.search(b"#!python", content, flags=re.MULTILINE) is not None

    if prefix is None:
        if sys.platform == "win32":  # unix: no cover
            prefix = os.path.dirname(sys.executable)
        else:
            prefix = os.path.dirname(os.path.dirname(sys.executable))

    if sys.platform == "win32":  # unix: no cover
        # On Windows a script present in a distribution might be installed
        # using different mechanisms:
        #
        # 1- the python file itself in the Scripts subdirectory
        # 2- a .exe with the same basename as the original python script which
        #    which call a python script called <basename>-script.py
        # 3- a .exe without a side python script
        script = (
            os.path.join(prefix, name)
            if os.path.basename(prefix) == "scripts"
            else os.path.join(prefix, "Scripts", name)
        )

        if script.endswith(".exe"):
            script_exe = script
            script_py = script[:-4] + "-script.py"
        else:
            script_exe = script + ".exe"
            script_py = script + "-script.py"

        if os.path.isfile(script_py):
            # If we have a side <basename>-script.py always use it, instead of
            # the .exe
            return [interpreter(prefix), script_py]
        elif os.path.isfile(script_exe):
            # A .exe without side python script
            if has_relative_python_shebang(script_exe):  # all: no cover
                # relocatable python distribution
                return [interpreter(prefix), script_exe]
            return [script_exe]
        else:
            # Case in which the script is probably a Python file
            return [interpreter(prefix), script]
    else:
        return [interpreter(prefix), os.path.join(prefix, "bin", name)]



from __future__ import annotations

from typing import TYPE_CHECKING
import e3.log
from e3.anod.action import (
    Build,
    BuildOrDownload,
    Checkout,
    CreateSource,
    CreateSourceOrDownload,
    CreateSources,
    Decision,
    DownloadBinary,
    DownloadSource,
    GetSource,
    Install,
    InstallSource,
    Root,
    Test,
    Upload,
    UploadBinaryComponent,
    UploadSource,
    UploadSourceComponent,
)
from e3.anod.deps import Dependency
from e3.anod.error import AnodError
from e3.anod.package import UnmanagedSourceBuilder
from e3.anod.spec import has_primitive
from e3.collection.dag import DAG
from e3.env import BaseEnv
from e3.error import E3Error

if TYPE_CHECKING:
    from typing import Callable, Dict, FrozenSet, List, NoReturn, Optional, Tuple, Union
    from e3.anod.action import Action
    from e3.anod.package import SourceBuilder
    from e3.anod.spec import Anod
    from e3.anod.loader import AnodSpecRepository
    from e3.anod.sandbox import SandBox
    from e3.collection.dag import VertexID
    from e3.platform import Platform

    # spec name, build env, target env, host env, qualifier, kind, source name
    CacheKeyType = Tuple[
        str, Platform, Platform, Platform, Optional[str], Optional[str], Optional[str]
    ]
    ResolverType = Callable[[Action, Decision], bool]

logger = e3.log.getLogger("anod.context")


class SchedulingError(E3Error):
    """Exception raised by scheduling algorithm."""

    def __init__(
        self,
        message: Union[str, List[str]],
        origin: Optional[str] = None,
        uid: Optional[VertexID] = None,
        initiators: Optional[List[VertexID]] = None,
    ):
        """Scheduling error initialization.

        :param message: the exception message
        :param origin: the name of the function, class, or module having raised
            the exception
        :param uid: uid of action that cause the error
        :param initiators: list of uids involved in the failure
        """
        super().__init__(message, origin)
        self.uid = uid
        self.initiators = initiators if initiators is not None else []


class AnodContext:
    """Anod context.

    :ivar repo: an anod spec repository
    :vartype repo: e3.anod.loader.AnodSpecRepository
    :ivar tree: a DAG containing the list of possible actions
    :ivar root: root node of the DAG
    :ivar cache: cache of anod instances, indexed by the spec's name.
    :vartype cache: dict[e3.anod.spec.Anod]
    :ivar sources: list of available sources in the current context,
        indexed by the source's name.
    :vartype sources: list[e3.anod.package.SourceBuilder]
    :ivar default_env: default environment (used to override build='default')
        when simulating a list of action from another machine.

    :ivar plan: maintain a link between a plan line and the generated actions
        which is useful for setting parameters such as weather or process that
        are conveyed by the plan and not by the specs
    """

    def __init__(
        self,
        spec_repository: AnodSpecRepository,
        default_env: Optional[BaseEnv] = None,
        reject_duplicates: bool = False,
    ):
        """Initialize a new context.

        :param spec_repository: an Anod repository
        :param default_env: an env that should be considered as the
            default for the current context. Mainly useful to simulate
            another server context. If None then we assume that the
            context if the local server
        :param reject_duplicates: if True, raise SchedulingError when two
            duplicated action are generated
        """
        self.repo = spec_repository
        if default_env is None:
            self.default_env = BaseEnv()
        else:
            self.default_env = default_env.copy()
        self.reject_duplicates = reject_duplicates

        self.tree = DAG()
        self.root = Root()
        self.dependencies: Dict[str, Dict[str, Tuple[Dependency, Anod]]] = {}
        self.add(self.root)
        self.cache: Dict[CacheKeyType, Anod] = {}
        self.sources: Dict[str, Tuple[str, SourceBuilder]] = {}

    def load(
        self,
        name: str,
        env: Optional[BaseEnv],
        qualifier: Optional[str],
        kind: str,
        sandbox: Optional[SandBox] = None,
        source_name: Optional[str] = None,
    ) -> Anod:
        """Load a spec instance.

        :param name: spec name
        :param env: environment to use for the spec instance
        :param qualifier: spec qualifier
        :param kind: primitive used for the loaded spec
        :param sandbox: is not None bind the anod instances to a sandbox
        :param source_name: when the primitive is "source" we create a specific
            instance for each source package we have to create.
        :return: a spec instance
        """
        if env is None:
            env = self.default_env

        # Key used for the spec instance cache
        key = (name, env.build, env.host, env.target, qualifier, kind, source_name)

        if key not in self.cache:
            # Spec is not in cache so create a new instance
            self.cache[key] = self.repo.load(name)(
                qualifier=qualifier, env=env, kind=kind
            )
            if sandbox is not None:
                self.cache[key].bind_to_sandbox(sandbox)

            # Update tracking of dependencies
            self.dependencies[self.cache[key].uid] = {}

            # Update the list of available sources. ??? Should be done
            # once per spec (and not once per spec instance). Need some
            # spec cleanup to achieve that ???
            source_builders = self.cache[key].source_pkg_build
            if source_builders is not None:
                for s in source_builders:
                    self.sources[s.name] = (name, s)

        return self.cache[key]

    def add(self, data: Action, *args: Action) -> None:
        """Add node to context tree.

        :param data: node data
        :param args: list of predecessors
        """
        preds = [k.uid for k in args]
        self.tree.update_vertex(data.uid, data, predecessors=preds, enable_checks=False)

    def add_decision(
        self,
        decision_class: Callable[..., Decision],
        root: Action,
        left: Action,
        right: Action,
    ) -> None:
        """Add a decision node.

        This create the following subtree inside the dag::

            root --> decision --> left
                              |-> right

        :param decision_class: Decision subclass to use
        :param root: parent node of the decision node
        :param left: left decision (child of Decision node)
        :param right: right decision (child of Decision node)
        """
        decision_action = decision_class(root, left, right)
        self.add(decision_action, left, right)
        self.connect(root, decision_action)

    def connect(self, action: Action, *args: Action) -> None:
        """Add predecessors to a node.

        :param action: parent node
        :param args: list of predecessors
        """
        preds = [k.uid for k in args]
        self.tree.update_vertex(action.uid, predecessors=preds, enable_checks=False)

    def __contains__(self, data: Action) -> bool:
        """Check if a given action is already in the internal DAG.

        :param data: an Action
        """
        return data.uid in self.tree

    def __getitem__(self, key: str) -> Action:
        """Retrieve action from the internal DAG based on its key.

        :param key: action uid
        :return: an Action
        """
        return self.tree[key]

    def predecessors(self, action: Action) -> List[Action]:
        """Retrieve predecessors of a given action.

        :param action: the parent action
        :return: the predecessor list
        """
        return [self[str(el)] for el in self.tree.get_predecessors(action.uid)]

    def link_to_plan(self, vertex_id: str, plan_line: str, plan_args: dict) -> None:
        """Tag the vertex with plan info.

        :param vertex_id: ID of the vertex
        :param plan_line: corresponding line:linenumber in the plan
        :param plan_args: action args after plan execution, taking into
            account plan context (such as with defaults(XXX):)
        """
        if self.reject_duplicates:
            previous_tag = self.tree.get_tag(vertex_id=vertex_id)
            if previous_tag and previous_tag["plan_line"] != plan_line:
                raise SchedulingError(
                    "entries {} and {} conflict because they result in "
                    "the same build space (id: {}). Check your "
                    "build_space_name property or your qualifiers".format(
                        previous_tag["plan_line"], plan_line, vertex_id
                    )
                )
        self.tree.add_tag(vertex_id, {"plan_line": plan_line, "plan_args": plan_args})

    def add_anod_action(
        self,
        name: str,
        env: BaseEnv,
        primitive: str,
        qualifier: Optional[str] = None,
        source_packages: Optional[List[str]] = None,
        upload: bool = True,
        plan_line: Optional[str] = None,
        plan_args: Optional[dict] = None,
        sandbox: Optional[SandBox] = None,
    ) -> Action:
        """Add an Anod action to the context.

        :param name: spec name
        :param env: spec environment
        :param primitive: spec primitive
        :param qualifier: qualifier
        :param source_packages: if not empty only create the specified list of
            source packages and not all source packages defined in the anod
            specification file
        :param upload: if True consider uploading to the store
        :param plan_line: corresponding line:linenumber in the plan
        :param plan_args: action args after plan execution, taking into
            account plan context (such as with defaults(XXX):)
        :return: the root added action
        """
        # First create the subtree for the spec
        result = self.add_spec(
            name,
            env,
            primitive,
            qualifier,
            source_packages=source_packages,
            plan_line=plan_line,
            plan_args=plan_args,
            sandbox=sandbox,
            upload=upload,
        )

        # Resulting subtree should be connected to the root node
        self.connect(self.root, result)

        # Ensure decision is set in case of explicit build or install
        if primitive == "build":
            build_action = None
            for el in self.predecessors(result):
                if isinstance(el, BuildOrDownload):
                    el.set_decision(BuildOrDownload.BUILD, plan_line)
                    build_action = self[el.left]
            if build_action is None and isinstance(result, Build):
                build_action = result

            # Create upload nodes
            if build_action is not None:
                spec = build_action.data
                if spec.component is not None and upload:
                    upload_bin: Union[UploadBinaryComponent, UploadSourceComponent]
                    if spec.has_package:
                        upload_bin = UploadBinaryComponent(spec)
                    else:
                        upload_bin = UploadSourceComponent(spec)
                    self.add(upload_bin)
                    # ??? is it needed?
                    if plan_line is not None and plan_args is not None:
                        self.link_to_plan(
                            vertex_id=upload_bin.uid,
                            plan_line=plan_line,
                            plan_args=plan_args,
                        )
                    self.connect(self.root, upload_bin)
                    self.connect(upload_bin, build_action)

        elif primitive == "install":
            for el in self.predecessors(result):
                if isinstance(el, BuildOrDownload):
                    el.set_decision(BuildOrDownload.INSTALL, plan_line)
        return result

    def add_spec(
        self,
        name: str,
        env: BaseEnv,
        primitive: str,
        qualifier: Optional[str] = None,
        source_packages: Optional[List[str]] = None,
        expand_build: bool = True,
        source_name: Optional[str] = None,
        plan_line: Optional[str] = None,
        plan_args: Optional[dict] = None,
        sandbox: Optional[SandBox] = None,
        upload: bool = False,
    ) -> Union[Build, CreateSources, CreateSource, Install, Test]:
        """Expand an anod action into a tree (internal).

        :param name: spec name
        :param env: spec environment
        :param primitive: spec primitive
        :param qualifier: qualifier
        :param source_packages: if not empty only create the specified list of
            source packages and not all source packages defined in the anod
            specification file
        :param expand_build: should build primitive be expanded
        :param source_name: source name associated with the source
            primitive
        :param plan_line: corresponding line:linenumber in the plan
        :param plan_args: action args after plan execution, taking into
            account plan context (such as with defaults(XXX):)
        :param sandbox: if not None, anod instance are automatically bind to
            the given sandbox
        :param upload: if True consider uploads to the store (sources and
            binaries)
        """

        def add_action(data: Action, connect_with: Optional[Action] = None) -> None:
            self.add(data)
            if connect_with is not None:
                self.connect(connect_with, data)

        def add_dep(spec_instance: Anod, dep: Dependency, dep_instance: Anod) -> None:
            """Add a new dependency in an Anod instance dependencies dict.

            :param spec_instance: an Anod instance
            :param dep: the dependency we want to add
            :param dep_instance: the Anod instance loaded for that dependency
            """
            if dep.local_name in spec_instance.deps:
                raise AnodError(
                    origin="expand_spec",
                    message="The spec {} has two dependencies with the same "
                    "local_name attribute ({})".format(
                        spec_instance.name, dep.local_name
                    ),
                )
            spec_instance.deps[dep.local_name] = dep_instance

        # Initialize a spec instance
        e3.log.debug(
            "add spec: name:{}, qualifier:{}, primitive:{}".format(
                name, qualifier, primitive
            )
        )
        spec = self.load(
            name,
            qualifier=qualifier,
            env=env,
            kind=primitive,
            sandbox=sandbox,
            source_name=source_name,
        )
        result: Union[Build, CreateSources, CreateSource, Install, Test]

        # Initialize the resulting action based on the primitive name
        if primitive == "source":
            if source_name is not None:
                result = CreateSource(spec, source_name)

            else:
                # Create the root node
                result = CreateSources(spec)

                # A consequence of calling add_action here
                # will result in skipping dependencies parsing.
                add_action(result)

                if TYPE_CHECKING:
                    # When creating sources we know that the
                    # source_pkg_build attribute is set
                    assert spec.source_pkg_build is not None

                # Then one node for each source package
                for sb in spec.source_pkg_build:
                    if source_packages and sb.name not in source_packages:
                        # This source package is defined in the spec but
                        # explicitly excluded in the plan
                        continue
                    if isinstance(sb, UnmanagedSourceBuilder):
                        # do not create source package for unmanaged source
                        continue
                    sub_result = self.add_spec(
                        name=name,
                        env=env,
                        primitive="source",
                        source_name=sb.name,
                        plan_line=plan_line,
                        plan_args=plan_args,
                        sandbox=sandbox,
                        upload=upload,
                    )
                    self.connect(result, sub_result)

        elif primitive == "build":
            result = Build(spec)
        elif primitive == "test":
            result = Test(spec)
        elif primitive == "install":
            result = Install(spec)
        else:  # defensive code
            raise ValueError("add_spec error: %s is not known" % primitive)

        # If this action is directly linked with a plan line make sure
        # to register the link between the action and the plan even
        # if the action has already been added via another dependency
        if plan_line is not None and plan_args is not None:
            self.link_to_plan(
                vertex_id=result.uid, plan_line=plan_line, plan_args=plan_args
            )

        if (
            primitive == "install"
            and not spec.has_package
            and has_primitive(spec, "build")
        ):
            if plan_line is not None and plan_args is not None:
                # We have an explicit call to install() in the plan but the
                # spec has no binary package to download.
                raise SchedulingError(
                    "error in plan at {}: "
                    "install should be replaced by build".format(plan_line)
                )
            # Case in which we have an install dependency but no install
            # primitive. In that case the real dependency is a build tree
            # dependency. In case there is no build primitive and no
            # package keep the install primitive (usually this means there
            # is an overloaded download procedure).
            return self.add_spec(
                name,
                env,
                "build",
                qualifier,
                expand_build=False,
                plan_args=plan_args,
                plan_line=plan_line,
                sandbox=sandbox,
                upload=upload,
            )

        if expand_build and primitive == "build" and spec.has_package:
            # A build primitive is required and the spec defined a binary
            # package. In that case the implicit post action of the build
            # will be a call to the install primitive
            return self.add_spec(
                name,
                env,
                "install",
                qualifier,
                plan_args=None,
                plan_line=plan_line,
                sandbox=sandbox,
                upload=upload,
            )

        # Add this stage if the action is already in the DAG, then it has
        # already been added.
        if result in self:
            return result

        if not has_primitive(spec, primitive):
            raise SchedulingError(f"spec {name} does not support primitive {primitive}")

        # Add the action in the DAG
        add_action(result)

        if primitive == "install":
            # Expand an install node to
            #    install --> decision --> build
            #                         \-> download binary
            download_action = DownloadBinary(spec)
            add_action(download_action)

            if has_primitive(spec, "build"):
                build_action = self.add_spec(
                    name=name,
                    env=env,
                    primitive="build",
                    qualifier=qualifier,
                    expand_build=False,
                    plan_args=None,
                    plan_line=plan_line,
                    sandbox=sandbox,
                    upload=upload,
                )
                self.add_decision(
                    BuildOrDownload, result, build_action, download_action
                )
            else:
                self.connect(result, download_action)

        elif primitive == "source":
            if source_name is not None:
                # Also add an UploadSource action
                if upload:
                    upload_src = UploadSource(spec, source_name)
                    self.add(upload_src)
                    # Link the upload to the current context
                    if plan_line is not None and plan_args is not None:
                        self.link_to_plan(
                            vertex_id=upload_src.uid,
                            plan_line=plan_line,
                            plan_args=plan_args,
                        )

                    self.connect(self.root, upload_src)
                    self.connect(upload_src, result)

                if TYPE_CHECKING:
                    # When creating sources we know that the
                    # source_pkg_build attribute is set
                    assert spec.source_pkg_build is not None
                for sb in spec.source_pkg_build:
                    if sb.name == source_name:
                        for checkout in sb.checkout:
                            if checkout not in self.repo.repos:
                                raise SchedulingError(
                                    origin="add_spec",
                                    message=f"unknown repository {checkout}",
                                )
                            co = Checkout(checkout, self.repo.repos[checkout])
                            add_action(co, result)

        # Look for dependencies
        spec_dependencies = []
        if (
            "%s_deps" % primitive in dir(spec)
            and getattr(spec, "%s_deps" % primitive) is not None
        ):
            spec_dependencies += getattr(spec, "%s_deps" % primitive)

        for e in spec_dependencies:
            if isinstance(e, Dependency):
                if e.kind == "source":
                    # A source dependency does not create a new node but
                    # ensure that sources associated with it are available
                    child_instance = self.load(
                        e.name,
                        kind="source",
                        env=self.default_env,
                        qualifier=None,
                        sandbox=sandbox,
                    )
                    add_dep(spec_instance=spec, dep=e, dep_instance=child_instance)
                    self.dependencies[spec.uid][e.local_name] = (
                        e,
                        spec.deps[e.local_name],
                    )

                    continue

                child_action = self.add_spec(
                    name=e.name,
                    env=e.env(spec, self.default_env),
                    primitive=e.kind,
                    qualifier=e.qualifier,
                    plan_args=None,
                    plan_line=plan_line,
                    sandbox=sandbox,
                    upload=upload,
                )

                add_dep(
                    spec_instance=spec, dep=e, dep_instance=child_action.anod_instance
                )
                self.dependencies[spec.uid][e.local_name] = (e, spec.deps[e.local_name])

                if e.kind == "build" and self[child_action.uid].data.kind == "install":
                    # We have a build tree dependency that produced a
                    # subtree starting with an install node. In that case
                    # we expect the user to choose BUILD as decision.
                    dec = self.predecessors(child_action)[0]
                    if isinstance(dec, BuildOrDownload):
                        dec.add_trigger(
                            result,
                            BuildOrDownload.BUILD,
                            plan_line if plan_line is not None else "unknown line",
                        )

                # Connect child dependency
                self.connect(result, child_action)

        # Look for source dependencies (i.e sources needed)
        if "%s_source_list" % primitive in dir(spec):
            source_list = getattr(spec, f"{primitive}_source_list")
            for s in source_list:
                # set source builder
                if s.name in self.sources:
                    s.set_builder(self.sources[s.name])
                # set other sources to compute source ignore
                s.set_other_sources(source_list)
                # add source install node
                src_install_uid = (
                    result.uid.rsplit(".", 1)[0] + ".source_install." + s.name
                )
                src_install_action = InstallSource(src_install_uid, spec, s)
                add_action(src_install_action, connect_with=result)

                # Then add nodes to create that source (download or creation
                # using anod source and checkouts)
                if s.name in self.sources:
                    spec_decl, obj = self.sources[s.name]
                else:
                    raise AnodError(
                        origin="expand_spec",
                        message="source %s does not exist "
                        "(referenced by %s)" % (s.name, result.uid),
                    )

                src_get_action = GetSource(obj)
                if src_get_action in self:
                    self.connect(src_install_action, src_get_action)
                    continue

                add_action(src_get_action, connect_with=src_install_action)

                src_download_action = DownloadSource(obj)
                add_action(src_download_action)

                if isinstance(obj, UnmanagedSourceBuilder):
                    # In that case only download is available
                    self.connect(src_get_action, src_download_action)
                else:
                    source_action = self.add_spec(
                        name=spec_decl,
                        env=self.default_env,
                        primitive="source",
                        plan_args=None,
                        plan_line=plan_line,
                        source_name=s.name,
                        sandbox=sandbox,
                        upload=upload,
                    )
                    for repo in obj.checkout:
                        r = Checkout(repo, self.repo.repos[repo])
                        add_action(r, connect_with=source_action)
                    self.add_decision(
                        CreateSourceOrDownload,
                        src_get_action,
                        source_action,
                        src_download_action,
                    )
        return result

    @classmethod
    def decision_error(cls, action: Action, decision: Decision) -> NoReturn:
        """Raise SchedulingError.

        :param action: action to consider
        :param decision: decision to resolve
        :raise SchedulingError
        """
        if decision.choice is None and decision.expected_choice in (
            Decision.LEFT,
            Decision.RIGHT,
        ):

            if decision.expected_choice == BuildOrDownload.BUILD:
                msg = (
                    "A spec in the plan has a build_tree dependency"
                    " on {spec}. Either explicitly add the line {plan_line}"
                    " or change the dependency to set"
                    ' require="installation" if possible'.format(
                        spec=action.data.name,
                        plan_line=decision.suggest_plan_fix(decision.expected_choice),
                    )
                )
            else:
                msg = "This plan resolver requires an explicit {}".format(
                    decision.suggest_plan_fix(decision.expected_choice)
                )
        elif decision.choice is None and decision.expected_choice is None:
            left_decision = decision.suggest_plan_fix(Decision.LEFT)
            right_decision = decision.suggest_plan_fix(Decision.RIGHT)
            msg = (
                "This plan resolver cannot decide whether what to do for"
                " resolving {}.".format(decision.initiator)
            )
            if left_decision is not None and right_decision is not None:
                msg += " Please either add {} or {} in the plan".format(
                    left_decision, right_decision
                )
        elif decision.choice == Decision.BOTH:
            msg = f"cannot do both {decision.left} and {decision.right}"
        else:
            trigger_decisions = "\n".join(
                "{} made by {} initiated by {}".format(
                    decision.left
                    if trigger_decision == Decision.LEFT
                    else decision.right,
                    trigger_action,
                    trigger_plan_line,
                )
                for (
                    trigger_action,
                    trigger_decision,
                    trigger_plan_line,
                ) in decision.triggers
            )
            conflict_choice = decision.choice
            if TYPE_CHECKING:
                # we expect the decision to be either LEFT or RIGHT
                # at this stage
                assert conflict_choice is not None
            msg = (
                "explicit {} decision made by {} conflicts with the "
                "following decision{}:\n{}".format(
                    decision.description(conflict_choice),
                    decision.decision_maker,
                    "s" if len(decision.triggers) > 1 else "",
                    trigger_decisions,
                )
            )

        raise SchedulingError(msg)

    @classmethod
    def always_download_source_resolver(
        cls, action: Action, decision: Decision
    ) -> bool:
        """Force source download when scheduling a plan.

        The resolver takes the following decision:
        * sources are always downloaded
        * any build that produces a package should be added explicitly

        :param action: action to consider
        :param decision: decision to resolve
        :return: True if the action should be scheduled, False otherwise
        :raise SchedulingError: in case no decision can be taken
        """
        if isinstance(action, CreateSource):
            return False
        elif isinstance(action, DownloadSource):
            return True
        else:
            return cls.decision_error(action, decision)

    @classmethod
    def always_create_source_resolver(cls, action, decision):
        """Force source creation when scheduling a plan."""
        if isinstance(action, CreateSource):
            return True
        elif isinstance(action, DownloadSource):
            return False
        else:
            return cls.decision_error(action, decision)

    def schedule(self, resolver: ResolverType) -> DAG:
        """Compute a DAG of scheduled actions.

        :param resolver: a function that helps the scheduler resolve cases
            for which a decision should be taken
        """
        rev = self.tree.reverse_graph()
        uploads: List[Tuple[Upload, FrozenSet[VertexID]]] = []
        dag = DAG()

        # Retrieve existing tags
        dag.tags = self.tree.tags

        # Note that schedule perform a pruning on the DAG, thus no cycle can
        # be introduced. That's why checks are disabled when creating the
        # result graph.
        for uid, action in rev:
            if TYPE_CHECKING:
                assert uid is not None
            if uid == "root":
                # Root node is always in the final DAG
                dag.update_vertex(uid, action, enable_checks=False)
            elif isinstance(action, Decision):
                # Decision node does not appears in the final DAG but we need
                # to apply the triggers based on the current list of scheduled
                # actions.
                action.apply_triggers(dag)
            elif isinstance(action, Upload):
                uploads.append((action, self.tree.get_predecessors(uid)))
            else:
                if TYPE_CHECKING:
                    assert isinstance(action, Action)
                # Compute the list of successors for the current node (i.e:
                # predecessors in the reversed graph). Ignore Upload
                # nodes as they will be processed only once the scheduling
                # is done.
                preds = [
                    k
                    for k in rev.get_predecessors(uid)
                    if not isinstance(rev[k], Upload)
                ]

                if len(preds) == 1 and isinstance(rev[preds[0]], Decision):
                    decision = rev[preds[0]]
                    # The current node addition is driven by a decision

                    # First check that the parent of the decision is
                    # scheduled. If not discard the item.
                    if decision.initiator not in dag:
                        continue

                    # Now check decision made. If the decision cannot be made
                    # delegate to the resolve function.
                    choice = decision.get_decision()

                    if choice == uid:
                        dag.update_vertex(uid, action, enable_checks=False)
                        dag.update_vertex(
                            decision.initiator, predecessors=[uid], enable_checks=False
                        )
                    elif choice is None:
                        # delegate to resolver
                        try:
                            if resolver(action, decision):
                                dag.update_vertex(uid, action, enable_checks=False)
                                dag.update_vertex(
                                    decision.initiator,
                                    predecessors=[uid],
                                    enable_checks=False,
                                )
                        except SchedulingError as e:
                            # In order to help the analysis of a scheduling
                            # error compute the explicit initiators of that
                            # action
                            dag.update_vertex(uid, action, enable_checks=False)
                            dag.update_vertex(
                                decision.initiator,
                                predecessors=[action.uid],
                                enable_checks=False,
                            )
                            rev_graph = dag.reverse_graph()
                            # Initiators are explicit actions (connected to
                            # 'root') that are in the closure of the failing
                            # node.
                            initiators = [
                                iuid
                                for iuid in rev_graph.get_closure(uid)
                                if "root" in rev_graph.get_predecessors(iuid)
                            ]
                            raise SchedulingError(
                                e.messages, uid=uid, initiators=initiators
                            )
                else:
                    # An action is scheduled only if one of its successors is
                    # scheduled.
                    successors = [k for k in preds if k in dag]
                    if successors:
                        dag.update_vertex(uid, action, enable_checks=False)
                        for a in successors:
                            dag.update_vertex(
                                a, predecessors=[uid], enable_checks=False
                            )

        # Handle Upload nodes. Add the node only if all predecessors
        # are scheduled.
        for action, predecessors in uploads:
            if len([p for p in predecessors if p not in dag]) == 0:
                dag.update_vertex(
                    action.uid, action, predecessors=predecessors, enable_checks=False
                )
                # connect upload to the root node
                dag.update_vertex(
                    "root", predecessors=[action.uid], enable_checks=False
                )
        return dag



import json
import os

from e3.env import Env
from e3.error import E3Error
from e3.fingerprint import Fingerprint

import pytest


def test_fingerprint():
    f1 = Fingerprint()
    f1.add("foo", "2")

    f2 = Fingerprint()
    f2.add("foo", "4")

    f12_diff = f2.compare_to(f1)
    assert f12_diff["new"] == set()
    assert f12_diff["updated"] == {"foo"}
    assert f12_diff["obsolete"] == set()

    f3 = Fingerprint()
    f3.add_file(__file__)

    f23_diff = f3.compare_to(f2)
    assert f23_diff["new"] == {"foo"}
    assert f23_diff["updated"] == set()
    assert f23_diff["obsolete"] == {os.path.basename(__file__)}

    assert f1.checksum() != f2.checksum() != f3.checksum()

    assert Env().build.os.version in str(f3)

    f4 = Fingerprint()
    f4.add_file(__file__)
    assert f4 == f3

    f5 = Fingerprint()
    with pytest.raises(E3Error) as err:
        f5.add("f4", f4)
    assert "f4 should be a string" in str(err.value)

    f6 = Fingerprint()
    f6.add("unicode", "6")
    assert len(f6.checksum()) == 64


def test_add_order_not_important():
    def create_fingerprint(first, second, third):
        """Create a fingerprint with 3 elements added in the given order.

        The arguments, first, second, and third should be either
        1, 2, or 3 (integers), and are meant as indices to (key, value)
        pairs in the idx_to_entry_map below, indicating which part
        of the fingerprint gets added in what order.
        """
        idx_to_entry_map = {1: ("foo", "1"), 2: ("bar", "2"), 3: ("baz", "3")}
        f = Fingerprint()
        f.add(*idx_to_entry_map[first])
        f.add(*idx_to_entry_map[second])
        f.add(*idx_to_entry_map[third])
        return f

    f_ref = create_fingerprint(1, 2, 3)

    def check_scenario(first, second, third):
        """Check scenario where adding fingerprint elements in given order.

        Create a fingerprint using create_fingerprint above
        (same arguments), and then check the resulting key against
        the reference key f_ref above.  Both the reference fingerprint
        and the new fingerprint have the same contents, so they should
        behave the same in all ways.
        """
        f = create_fingerprint(first, second, third)
        assert f == f_ref
        assert f.compare_to(f_ref) is None
        assert f_ref.compare_to(f) is None
        assert str(f) == str(f_ref)
        assert f.checksum() == f_ref.checksum()

    check_scenario(1, 2, 3)
    check_scenario(1, 3, 2)
    check_scenario(2, 1, 3)
    check_scenario(2, 3, 1)
    check_scenario(3, 1, 2)
    check_scenario(3, 2, 1)


def test_fingerprint_version():
    """Changing the FINGERPRINT_VERSION modify the fingerprint's checksum."""
    import e3.fingerprint

    f1 = Fingerprint()

    e3.fingerprint.FINGERPRINT_VERSION = "0.0"
    f2 = Fingerprint()

    assert f1 != f2
    assert f1.checksum() != f2.checksum()

    f3 = Fingerprint()

    assert f2 == f3
    assert f2.checksum() == f3.checksum()


def test_invalid_fingerprint():
    """A fingerprint value should be hashable."""
    f1 = Fingerprint()
    with pytest.raises(E3Error):
        f1.add("invalid", {})


def test_fingerprint_eq():
    """Check fingerprint __eq__ function."""
    f1 = Fingerprint()
    f1.add("1", "1")
    assert f1 != 1

    f2 = Fingerprint()
    f2.add("1", "1")
    f2.add("2", "2")
    assert f1 != f2
    assert f1.checksum() != f2.checksum()

    assert f1.compare_to(f1) is None


def test_fingerprint_save_and_load():
    # Create a directory where to store our fingerprints, allowing us
    # to use any fingerprint name without potentially colliding with
    # other files used by this testcase.
    os.mkdir("fingerprints")

    def fingerprint_path(filename):
        return os.path.join("fingerprints", filename)

    # Save and then load a minimal fingerprint...
    f_min = Fingerprint()

    f_min_filename = fingerprint_path("f_min")
    assert not os.path.exists(f_min_filename)
    f_min.save_to_file(f_min_filename)

    f_min_restored = Fingerprint.load_from_file(f_min_filename)
    assert f_min_restored == f_min
    assert str(f_min_restored) == str(f_min)
    assert f_min_restored.checksum() == f_min.checksum()

    # Save and then load a fingerprint with more data than the minimum.

    f2 = Fingerprint()
    f2.add("job1", "job1sha1")
    f2.add("job2", "sha1job2")

    f2_filename = fingerprint_path("f2")
    assert not os.path.exists(f2_filename)
    f2.save_to_file(f2_filename)

    f2_restored = Fingerprint.load_from_file(f2_filename)
    assert f2_restored == f2
    assert str(f2_restored) == str(f2)
    assert f2_restored.checksum() == f2.checksum()

    # Trying to load from a file with invalid contents (bad JSON)

    f_bad_filename = fingerprint_path("f_bad_JSON")
    with open(f_bad_filename, "w") as f:
        f.write("yello{")

    f3 = Fingerprint.load_from_file(f_bad_filename)
    assert f3 is None

    # Trying to load from a file which contains valid data, but
    # is not an dictionary, and therefore clearly not something
    # that comes from a fingerprint...

    f_not_filename = fingerprint_path("not_a_fingerprint")
    with open(f_not_filename, "w") as f:
        json.dump([1, 2, 3], f)

    f4 = Fingerprint.load_from_file(f_not_filename)
    assert f4 is None

    # Try to load from a file which is missing one of the mandatory
    # elements.

    for key in ("fingerprint_version", "elements"):
        f_key_missing_filename = fingerprint_path("no_%s_key")

        # To create the bad file without assuming too much in this test
        # how the fingerprint is saved to file, we save a valid fingerprint
        # to file, load that file back, remove the key, and then save the
        # truncated data again.

        f2.save_to_file(f_key_missing_filename)
        with open(f_key_missing_filename) as f:
            data = json.load(f)
        del data[key]
        with open(f_key_missing_filename, "w") as f:
            json.dump(data, f)

        f5 = Fingerprint.load_from_file(f_key_missing_filename)
        assert f5 is None

    # Try loading a fingerprint whose version number is not recognized
    # (typically, and old fingerprint version that we no longer support).
    #
    # To create the file without assuming too much in this test how
    # fingerprint are saved to file, we start with a valid fingerprint
    # that we saved to a file, load that file back, adjust the version
    # number, and then replace the good fingeprint in that file by
    # the modified one.

    f_bad_version = fingerprint_path("bad_version")

    f2.save_to_file(f_bad_version)
    with open(f_bad_version) as f:
        data = json.load(f)
        data["fingerprint_version"] = "1.0"
        data["elements"]["fingerprint_version"] = "1.0"
    with open(f_bad_version, "w") as f:
        json.dump(data, f)

    f = Fingerprint.load_from_file(f_bad_version)
    assert f is None



import os

import e3.diff
import e3.fs

import pytest


def test_non_existing():
    """Check that a non existing file will be considered as null string."""
    assert e3.diff.diff("foo1", "foo2") == ""


def test_patch():
    test_dir = os.path.dirname(__file__)
    file_to_patch = os.path.join(test_dir, "file_to_patch.orig.txt")
    file_after_patch = os.path.join(test_dir, "file_to_patch.new.txt")
    file_after_patch2 = os.path.join(test_dir, "file_to_patch.new2.txt")
    file_patch = os.path.join(test_dir, "patch.txt")
    file_patch2 = os.path.join(test_dir, "patch2.txt")

    current_dir = os.getcwd()

    e3.fs.cp(file_to_patch, current_dir)
    e3.diff.patch(file_patch, current_dir)

    with open("file_to_patch.orig.txt") as fd:
        output = fd.readlines()

    with open(file_after_patch) as fd:
        expected = fd.readlines()

    # By default empty line, leading and trailing whitespaces are ignored
    assert e3.diff.diff(expected + [""], output) == ""
    assert e3.diff.diff([" " + line + " " for line in expected], output) == ""
    assert e3.diff.diff(expected, output, ignore_white_chars=False) == ""

    # we can also ignore specific pattern
    assert "-to ignore" in e3.diff.diff(expected + ["to ignore"], output)
    assert e3.diff.diff(expected + ["to ignore"], output, ignore="ig.*re") == ""

    e3.fs.cp(file_patch2, current_dir)
    e3.diff.patch("patch2.txt", current_dir, discarded_files=["dummy_to_patch.new.txt"])
    assert e3.diff.diff("file_to_patch.orig.txt", file_after_patch2) == ""

    e3.fs.cp(file_patch2, current_dir)
    e3.diff.patch(
        "patch2.txt", current_dir, discarded_files=lambda x: x.endswith("new.txt")
    )


def test_patch_ignore_all(caplog):
    test_dir = os.path.dirname(__file__)
    file_to_patch = os.path.join(test_dir, "file_to_patch.orig.txt")
    file_patch2 = os.path.join(test_dir, "patch2.txt")

    current_dir = os.getcwd()

    e3.fs.cp(file_to_patch, current_dir)
    e3.fs.cp(file_patch2, current_dir)
    e3.diff.patch("patch2.txt", current_dir, discarded_files=lambda x: True)

    assert "All patch2.txt content has been discarded" in caplog.text


def test_discarded():
    test_dir = os.path.dirname(__file__)
    orig = os.path.join(test_dir, "data.txt")
    new = os.path.join(test_dir, "data_new.txt")
    current_dir = os.getcwd()

    e3.fs.cp(orig, current_dir)
    e3.fs.cp(os.path.join(test_dir, "data_patch_universal.txt"), current_dir)
    e3.diff.patch(
        "data_patch_universal.txt", current_dir, discarded_files=["*file_to_patch*"]
    )

    with open(new) as fd:
        expected = fd.read()
    with open("data.txt") as fd:
        result = fd.read()
    assert result == expected

    e3.fs.cp(orig, current_dir)
    e3.fs.cp(os.path.join(test_dir, "data_patch_contextual.txt"), current_dir)
    e3.diff.patch(
        "data_patch_universal.txt", current_dir, discarded_files=["*file_to_patch*"]
    )

    with open(new) as fd:
        expected = fd.read()
    with open("data.txt") as fd:
        result = fd.read()
    assert result == expected

    e3.fs.cp(orig, current_dir)


def test_patch_invalid():
    test_dir = os.path.dirname(__file__)
    file_to_patch = os.path.join(test_dir, "file_to_patch.orig.txt")
    file_patch2 = os.path.join(test_dir, "patch2.txt")

    current_dir = os.getcwd()

    e3.fs.cp(file_to_patch, current_dir)
    e3.fs.cp(file_patch2, current_dir)

    with open("patch2.txt", "a") as f:
        f.write("invalid")

    with pytest.raises(e3.diff.DiffError):
        e3.diff.patch("patch2.txt", current_dir)



import base64
import hashlib
import json

from e3.net.token import get_payload, is_valid, utc_timestamp

FUTURE_TIMESTAMP = 9999999999999


def create_token(payload: dict) -> str:
    """Return JSON Web Token.

    :param payload:
    :return: token
    """
    payload_bytes = json.dumps(payload).encode("utf-8")
    data = (
        b"eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9" + b"." + base64.b64encode(payload_bytes)
    )
    signature = hashlib.sha256(data).hexdigest().encode("utf-8")
    token = data + b"." + signature
    return token.decode("utf-8")


def test_get_payload():
    expected = {
        "GivenName": "Toto",
        "Email": "toto@example.com",
        "Role": "tester",
        "iss": "adacore.com",
        "sub": "test",
        "exp": 1550655711,
        "iat": 1519119711,
        "Surname": "123",
        "aud": "www.adacore.com",
    }
    token = create_token(expected)
    payload = get_payload(token)
    assert payload == expected


def test_valid_token():
    valid_token = create_token({"typ": "Bearer", "exp": FUTURE_TIMESTAMP})
    assert is_valid(valid_token)

    near_future = utc_timestamp() + 7 * 60
    near_future_token = create_token({"typ": "Bearer", "exp": near_future})
    assert is_valid(near_future_token)


def test_wrong_token_type():
    badtype_token = create_token({"typ": "badtype", "exp": FUTURE_TIMESTAMP})
    assert not is_valid(badtype_token)


def test_old_token():
    old_token = create_token({"typ": "Bearer", "exp": 1419064452})
    assert not is_valid(old_token)

    # Verify that a token valid for less than 5 min will be considered invalid
    expire_soon_date = utc_timestamp() + 4 * 60
    expire_soon_token = create_token({"typ": "Bearer", "exp": expire_soon_date})
    assert not is_valid(expire_soon_token)


def test_exception_pass():
    assert not is_valid("..")


def test_token_str():
    token = (
        "eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJpc3MiOiJhZGFjb3JlLmNvbSI"
        "sImlhdCI6MTUxOTExOTcxMSwiZXhwIjoxNTUwNjU1NzExLCJhdWQiOiJ3d3cuYWR"
        "hY29yZS5jb20iLCJzdWIiOiJ0ZXN0IiwiR2l2ZW5OYW1lIjoiVG90byIsIlN1cm5"
        "hbWUiOiIxMjMiLCJFbWFpbCI6InRvdG9AZXhhbXBsZS5jb20iLCJSb2xlIjoidGV"
        "zdGVyIn0.UERDJuUFh6A30TXIEKbNu6BbDOUCdsLGMKuQS8DfboA"
    )
    payload = get_payload(token)
    expected = {
        "GivenName": "Toto",
        "Email": "toto@example.com",
        "Role": "tester",
        "iss": "adacore.com",
        "sub": "test",
        "exp": 1550655711,
        "iat": 1519119711,
        "Surname": "123",
        "aud": "www.adacore.com",
    }
    assert payload == expected
    assert not is_valid(token)



from __future__ import annotations

import os
from typing import TYPE_CHECKING

import e3.error
import e3.log
import e3.os.process
from e3.fs import mkdir, rm
from e3.os.fs import touch

logger = e3.log.getLogger("buildspace")


if TYPE_CHECKING:
    from typing import Optional, List


class BuildSpace:
    """Build space located inside a sandbox."""

    DIRS = (
        "binary",
        "build",
        "install",
        "log",
        "pkg",
        "results",
        "src",
        "test",
        "tmp",
        "src",
    )

    def __init__(self, root_dir: str):
        """Initialise a build space.

        :param root_dir: build space root dir
        """
        self.directory_mapping = {k: k for k in self.DIRS}
        self.root_dir = os.path.abspath(root_dir)
        self.initialized = False

    def exists(self) -> bool:
        """Return True if the build space exists on disk.

        This function just checks the contents of self.root_dir,
        and verifies that it appears to be a build space.
        It does so, irrespective of whether self.initialize
        is True or not.

        :return: True if self.root_dir is a buildspace, False otherwise.
        """
        # Start by verifying that the file used as build space markers
        # exists.
        if not os.path.isfile(os.path.join(self.root_dir, ".buildspace")):
            return False
        # Next, verify that all the necessary directories exist as well.
        for d in self.DIRS:
            if not os.path.isdir(self.subdir(name=d)):
                return False
        return True

    @property
    def dirs(self) -> List[str]:
        return list(self.directory_mapping.values())

    def subdir(self, name: str) -> str:
        if name not in self.DIRS:
            raise ValueError("invalid subdir %s" % name)
        return os.path.join(self.root_dir, self.directory_mapping[name])

    def __getattr__(self, name: str) -> str:
        if name.endswith("_dir") and name[:-4] in self.DIRS:
            return self.subdir(name[:-4])
        raise AttributeError("unknown build space attribute: %s" % name)

    def __getitem__(self, key: str) -> str:
        if key.isupper() and key.endswith("_DIR"):
            return getattr(self, key.lower(), None)
        raise KeyError("invalid build space key: %s" % key)

    def create(self, quiet: bool = False) -> None:
        """Create a build space.

        The function create all the necessary directories and files to have
        a valid build space. If the build space already exists, only tmp
        directory content is reset.

        :param quiet: do not print info messages
        """
        rm(self.tmp_dir, recursive=True)

        for d in self.DIRS:
            mkdir(self.subdir(name=d), quiet=quiet)

        # Add a marker that identify a build space
        touch(os.path.join(self.root_dir, ".buildspace"))

        self.initialized = True

    def reset(self, keep: Optional[List[str]] = None) -> None:
        """Reset build space.

        The function delete the complete buildspace. The only elements that
        are not deleted are the logs, the testsuite results and any
        subdirectories in keep parameter

        A call to self.create() is needed after calling this function.

        :param keep: a list of directory to keep in addition
            to results and log. Each element should be part of BuildSpace.DIRS
        """
        dirs_to_keep = set(keep) if keep is not None else set()
        dirs_to_keep.update(("results", "log"))

        dirs_to_reset = set(self.DIRS) - dirs_to_keep

        for d in dirs_to_reset:
            rm(self.subdir(name=d), recursive=True)
        self.initialized = False



"""Global environment and platform information support.

This package provide a class called Env used to store global
information. Env is a singleton so there is in fact only one instance.
"""

from __future__ import annotations

import abc
import os
import pickle
import sys
from collections import namedtuple

from typing import TYPE_CHECKING

import e3.log
import e3.os.platform
from e3.platform import Platform


if TYPE_CHECKING:
    from typing import Any, Dict, Iterable, List, Optional, Union
    from argparse import Namespace

logger = e3.log.getLogger("env")


# This global variable contains a list of tuples
# (build platform, host platform) that should not be considered as canadian
# configurations.
CANADIAN_EXCEPTIONS = (
    ("x86-windows", "x86_64-windows"),
    ("sparc-solaris", "sparc64-solaris"),
)


EnvInfo = namedtuple("EnvInfo", ["build", "host", "target"])


class AbstractBaseEnv(metaclass=abc.ABCMeta):
    """Environment Handling.

    Abstract class to factorize code between BaseEnv and Env.
    :ivar build: current build Platform
    :vartype build: Platform
    :ivar host: current host Platform
    :vartype host: Platform
    :ivar target: current target Platform
    :vartype target: Platform
    :ivar main_options: The command-line switches, after parsing by
    the e3.Main class (see the documentation of that class).
    """

    @abc.abstractmethod
    def __init__(
        self,
        build: Optional[Platform] = None,
        host: Optional[Platform] = None,
        target: Optional[Platform] = None,
    ):
        if not self._initialized:
            self.build = Platform.get() if build is None else build
            self.host = self.build if host is None else host
            self.target = self.host if target is None else target
            self.environ: Optional[dict] = None
            self.cwd: Optional[str] = None
            self.main_options: Optional[Namespace] = None

    @abc.abstractproperty
    def _initialized(self) -> bool:
        """Whether the new instance should be initialized.

        This is mostly useful to implement a singleton, as done in Env()
        """
        pass  # all: no cover

    @abc.abstractmethod
    def _items(self) -> Iterable:
        """Return the list of instance variables."""
        pass  # all: no cover

    @property
    def platform(self) -> str:
        """Compute the platform name based on the host and the target.

        For example for target ppc-elf hosted on linux, platform will
        be ppc-elf-linux. So the concept of platform embed both target
        and host concept.
        """
        if self.is_cross:
            # In cross we need to append host information. For backward
            # compatibility we don't append 64 to darwin host (which is
            # always 64bits).
            suffix = self.host.os.name
            if self.host.cpu.bits == 64 and self.host.os.name != "darwin":
                suffix += "64"
            return self.target.platform + "-" + suffix
        else:
            # In native concept the platform is equivalent to target.platform
            return self.target.platform

    @property
    def is_canadian(self) -> bool:
        """Return true if this is a canadian configuration."""
        if self.build != self.host:
            if (self.build.platform, self.host.platform) in CANADIAN_EXCEPTIONS:
                return False
            return True
        else:
            return False

    @property
    def is_cross(self) -> bool:
        """Return true if this is a cross configuration."""
        if self.target != self.host:
            return True
        else:
            return False

    def set_build(
        self,
        name: Optional[str] = None,
        version: Optional[str] = None,
        machine: Optional[str] = None,
        mode: Optional[str] = None,
    ) -> None:
        """Set build platform.

        :param name: a string that identify the system to be considered
            as the build. If None then build is unchanged. Note that passing
            an empty value will force the autodetection and possibly reset to
            the default value.
        :param version: a string containing the system version. If set
            to None the version is either a default or autodetected when
            possible
        :param machine: a string containing the name of the target
            machine.
        :param mode: a string containing the name of the mode. This
            notion is needed on some targets such as VxWorks to switch between
            kernel mode and other modes such as rtp

        When calling set_build, the target and host systems are reset to the
        build one. Thus you should call set_build before calling either
        set_host or set_target.
        """
        e3.log.debug("set_build (build_name=%s, build_version=%s)", name, version)
        self.build = Platform.get(
            platform_name=name, version=version, machine=machine, mode=mode
        )
        self.host = self.build
        self.target = self.build

    def set_host(
        self,
        name: Optional[str] = None,
        version: Optional[str] = None,
        machine: Optional[str] = None,
        mode: Optional[str] = None,
    ) -> None:
        """Set host platform.

        :param name: a string that identify the system to be considered
            as the host. If None then host is set to the build one (the
            autodetected platform). If set to 'build' or 'target' then host
            is set respectively to current 'build' or 'target' value
        :param version: a string containing the system version. If set to
            None the version is either a default or autodetected when possible
        :param machine: a string containing the name of the target
            machine.
        :param mode: a string containing the name of the mode. This
            notion is needed on some targets such as VxWorks to switch between
            kernel mode and other modes such as rtp

        When calling set_host, the target system is reset to the host one.
        Thus you should call set_host before set_target otherwise your call
        to set_target will be ignored. Note also that is the host_name is
        equal to the build platform, host_version will be ignored.
        """
        if name is None:
            name = "build"

        if name == "target":
            self.host = self.target
        elif name == "build":
            self.host = self.build
        else:
            self.host = Platform.get(
                platform_name=name, version=version, machine=machine, mode=mode
            )
        self.target = self.host

    def set_target(
        self,
        name: Optional[str] = None,
        version: Optional[str] = None,
        machine: Optional[str] = None,
        mode: Optional[str] = None,
    ) -> None:
        """Set target platform.

        :param name: a string that identify the system to be considered
            as the host. If None then host is set to the host one. If set to
            'build' or 'host' then target is set respectively to current
            'build' or 'host' value. In that case target_version and
            target_machine are ignored.
        :param version: a string containing the system version. If set
            to None the version is either a default or autodetected when
            possible.
        :param machine: a string containing the name of the target
            machine.
        :param mode: a string containing the name of the mode. This
            notion is needed on some targets such as VxWorks to switch between
            kernel mode and other modes such as rtp

        The target parameters are ignored if the target_name is equal to the
        host platform.
        """
        if name is None:
            name = "host"

        if name == "host":
            self.target = self.host
        elif name == "build":
            self.target = self.build
        else:
            self.target = Platform.get(
                platform_name=name, version=version, machine=machine, mode=mode
            )

    def set_env(
        self,
        build: Optional[str] = None,
        host: Optional[str] = None,
        target: Optional[str] = None,
    ) -> None:
        """Set build/host/target.

        :param build: string as passed to --build option
        :param host: string as passed to --host
        :param target: string as passed to --target
        """
        saved_build = self.build
        saved_host = self.host
        saved_target = self.target

        def get_platform(
            value: Optional[str], propagate_build_info: bool = False
        ) -> Optional[Platform]:
            """Platform based on string value.

            :param value: a string representing a platform or None
            :param propagate_build_info: whether to propagate machine name
                and OS version if no machine name set
            """
            if value is None:
                return None

            # We expect 4 fields for build and host and target
            split_value = ([k if k else None for k in value.split(",")] + [None] * 4)[
                0:4
            ]

            if split_value[0] == "build":
                return saved_build
            elif split_value[0] == "host":
                return saved_host
            elif split_value[0] == "target":
                return saved_target
            elif not propagate_build_info:
                return Platform.get(*split_value)  # type: ignore
            else:

                # Propagate machine name and OS version if necessary
                if split_value[2] is None:
                    # No new machine name specified, reuse the current one
                    split_value[2] = saved_build.machine

                    # And if there is no OS version set also keep the
                    # current one, setting build='x86-linux' on a 64bit
                    # Linux machine should not change the OS version
                    if split_value[1] is None:
                        split_value[1] = saved_build.os.version
                return Platform.get(*split_value)  # type: ignore

        # Retrieve final values for build, host and target
        build_opts = get_platform(build, propagate_build_info=True)
        host_opts = get_platform(host)
        target_opts = get_platform(target)

        # Apply new build, host and target in the right order
        if build_opts is not None:
            self.build = build_opts
            self.host = build_opts
            self.target = build_opts

        if host_opts is not None:
            self.host = host_opts
            self.target = host_opts

        if target_opts is not None:
            self.target = target_opts

    def str_triplet(self) -> EnvInfo:
        """Return a triplet of strings suitable to call set_env.

        :return: a namedtuple suitable for a call to set_env
        """
        result: List[Optional[str]] = []
        if not self.build.is_default:
            result.append(",".join([self.build.platform, self.build.os.version]))
        else:
            result.append(None)

        if self.host != self.build:
            result.append(",".join([self.host.platform, self.host.os.version]))
        else:
            result.append(None)

        if self.target != self.host:
            result.append(
                ",".join(
                    [
                        self.target.platform,
                        self.target.os.version,
                        self.target.machine,
                        self.target.os.mode,
                    ]
                )
            )
        else:
            result.append(None)
        return EnvInfo(*result)

    def cmd_triplet(self) -> List[str]:
        """Return command line parameters corresponding to current env.

        :return: a list of command line parameters
        """
        build_str, host_str, target_str = self.str_triplet()
        result = []
        if build_str is not None:
            result.append("--build=%s" % build_str)

        if host_str is not None:
            result.append("--host=%s" % host_str)

        if target_str is not None:
            result.append("--target=%s" % target_str)

        return result

    def get_attr(
        self, name: str, default_value: Any = None, forced_value: Any = None
    ) -> Any:
        """Return an attribute value.

        :param name: name of the attribute to check. Name can contain '.'
        :param default_value: returned value if forced_value not set and the
            attribute does not exist
        :param forced_value: if not None, this is the return value

        :return: the attribute value

        This function is useful to get the value of optional functions
        parameters whose default value might depend on the environment.
        """
        if forced_value is not None:
            return forced_value

        attributes = name.split(".")
        result: Any = self
        for a in attributes:
            if not hasattr(result, a):
                return default_value
            else:
                result = getattr(result, a)

        if result is None or result == "":
            return default_value

        return result

    @classmethod
    def add_path(cls, path: str, append: bool = False) -> None:
        """Set a path to PATH environment variable.

        :param path: path to add
        :param append: if True append, otherwise prepend. Default is prepend
        """
        cls.add_search_path("PATH", path, append)

    @classmethod
    def add_search_path(cls, env_var: str, path: str, append: bool = False) -> None:
        """Add a path to the env_var search paths.

        :param env_var: the environment variable name (e.g. PYTHONPATH,
            LD_LIBRARY_PATH, ...)
        :param path: path to add
        :param append: if True append, otherwise prepend. Default is prepend
        """
        if env_var not in os.environ:
            logger.debug(f"export {env_var}={path}")
            os.environ[env_var] = path
        else:
            if append:
                new_path = os.path.pathsep + path
                logger.debug(
                    "export {env_var}=${env_var}{new_path}".format(
                        env_var=env_var, new_path=new_path
                    )
                )
                os.environ[env_var] += new_path
            else:
                new_path = path + os.path.pathsep + os.environ[env_var]
                logger.debug(
                    "export {env_var}={new_path}".format(
                        env_var=env_var, new_path=new_path
                    )
                )
                os.environ[env_var] = new_path

    @property
    def dll_path_var(self) -> str:
        env_var_name = {"windows": "PATH", "darwin": "DYLD_FALLBACK_LIBRARY_PATH"}
        return env_var_name.get(self.host.os.name.lower(), "LD_LIBRARY_PATH")

    def add_dll_path(self, path: str, append: bool = False) -> None:
        """Add a path to the dynamic libraries search paths.

        :param path: path to add
        :param append: if True append, otherwise prepend. Default is prepend
        """
        # On most platforms LD_LIBRARY_PATH is used. For others use:
        self.add_search_path(self.dll_path_var, path, append)

    @property
    def discriminants(self) -> List[str]:
        """Compute discriminants.

        :return: the list of discriminants associated with the current context
            (target, host, ...). This is mainly used for testsuites to ensure a
            coherent set of base discriminants.
        """
        discs = [
            self.target.platform,
            self.target.triplet,
            self.target.cpu.endian + "-endian",
            self.target.cpu.name,
            self.host.os.name + "-host",
        ]

        if self.target.os.is_bareboard:
            discs.append("bareboard")
        else:
            discs.extend(
                (
                    self.target.os.name,
                    self.target.os.name + "-" + self.target.os.version,
                )
            )
        if self.target.os.name.startswith("vxworks"):  # all: no cover
            discs.append("vxworks")
        if not self.is_cross:
            discs.append("native")
        discs.append("%dbits" % self.target.cpu.bits)
        if self.target.os.name.lower() == "windows":
            discs.append("NT")

        if (not self.is_cross and not self.is_canadian) and self.build.is_virtual:
            discs.append("virtual_machine")  # all: no cover

        return discs

    @property
    def tmp_dir(self) -> str:
        """Return current temporary directory.

        :return: a path

        The function looks for several variables ``TMPDIR``, ``TMP``
        and in case none of these variables are defined fallback on
        on ``/tmp``.
        """
        return os.environ.get("TMPDIR", os.environ.get("TMP", "/tmp"))

    def to_dict(self) -> dict:
        """Get current env as a dictionary.

        :return: the dictionary entries are all strings and thus the result
            can be used to format string. For example ``Env().target.os.name``
            will appear with the key ``target_os_name``, ...
        """
        result = {k: v for k, v in self._items()}
        result["is_canadian"] = self.is_canadian
        result["is_cross"] = self.is_cross
        result["platform"] = self.platform

        for c in ("host", "target", "build"):
            result.update({f"{c}_{k}": v for k, v in result[c].to_dict().items()})
            del result[c]
        return result

    @classmethod
    def from_platform_name(cls, platform: str) -> Optional[AbstractBaseEnv]:
        """Return a BaseEnv object from a platform name.

        That's the reverse of platform property
        """
        # Is it a native platform?
        found = False
        e = BaseEnv()
        try:
            # If it is a native then set_build will work
            e.set_build(platform)
            found = True
        except KeyError:
            # Check whether is this a cross, in that case the platform name is:
            # <target-platform>-<host os name>[64]
            target_name, host = platform.rsplit("-", 1)
            if host == "darwin":
                host_cpu = "x86_64"
            elif host == "solaris":
                host_cpu = "sparc"
            elif host.endswith("64"):
                host = host[:-2]
                host_cpu = "x86_64"
            else:
                host_cpu = "x86"
            try:
                e.set_build(f"{host_cpu}-{host}")
                e.set_target(target_name)
                found = True
            except KeyError:
                # invalid platform name
                pass
        if not found:
            return None
        else:
            # Verify that the computed platform is equal to what we had
            assert e.platform == platform
            return e


class BaseEnv(AbstractBaseEnv):
    """BaseEnv."""

    _initialized = False
    # Not a singleton, always initialize new instance

    def __init__(
        self,
        build: Optional[Platform] = None,
        host: Optional[Platform] = None,
        target: Optional[Platform] = None,
    ):
        """Initialize a BaseEnv object.

        On first instantiation, build attribute will be computed and host
        and target set to the build attribute.

        :param build: build architecture. If None then it is set to default
            build
        :param host: host architecture. If None then it is set to build
        :param target: target architecture. If None then it is set to target
        """
        # class variable that holds the current environment
        self._instance: Dict[str, Any] = {}

        # class variable that holds the stack of saved environments state
        self._context: List[Any] = []
        super().__init__(build, host, target)

    def __setattr__(self, name: str, value: Any) -> None:
        if name in ("_instance", "_context"):
            object.__setattr__(self, name, value)
        else:
            self._instance[name] = value

    def __getattr__(self, name: str) -> None:
        try:
            return self._instance[name]
        except KeyError as e:
            raise AttributeError(e).with_traceback(sys.exc_info()[2])

    def _items(self) -> Iterable[Any]:
        return iter(self._instance.items())

    def copy(
        self,
        build: Optional[str] = None,
        host: Optional[str] = None,
        target: Optional[str] = None,
    ) -> BaseEnv:
        """Copy an env.

        :param build: like build set_env parameter
        :param host: like host set_env parameter
        :param target: like target set_env parameter
        :return: a deep copy of the current env
        """
        result = BaseEnv()
        for k, v in self._items():
            setattr(result, k, v)
        result.set_env(build, host, target)
        return result

    @classmethod
    def from_env(cls, env: Optional[Union[Env, BaseEnv]] = None) -> BaseEnv:
        """Return a new BaseEnv object from an env.

        :param env: env. If None copy the current Env
        """
        if env is None:
            return BaseEnv(build=Env().build, host=Env().host, target=Env().target)
        else:
            return BaseEnv(build=env.build, host=env.host, target=env.target)


class Env(AbstractBaseEnv):
    """Env shows the current environment in used.

    Env is a singleton holding the current environment and platform
    information. It is set by e3.main when the --build/--host/--target option
    are passed to the command line and can be then changed by calling
    :meth:`set_build`, :meth:`set_host`, and :meth:`set_target`.
    """

    # class variable that holds the current environment
    _instance: Dict[str, Any] = {}

    # class variable that holds the stack of saved environments state
    _context: List[Any] = []

    def __init__(self) -> None:
        """Initialize or reuse an existing Env object (singleton).

        On first instantiation, build attribute will be computed and
        host and target set to the build attribute.
        """
        super().__init__()

    @property
    def _initialized(self) -> bool:
        return "build" in Env._instance

    def __setattr__(self, name: str, value: Any) -> None:
        if name == "_instance":
            Env._instance = value
        elif name == "_context":
            Env._context = value
        else:
            self._instance[name] = value

    def __getattr__(self, name: str) -> Any:
        try:
            return self._instance[name]
        except KeyError as e:
            raise AttributeError(e).with_traceback(sys.exc_info()[2])

    def _items(self) -> Iterable[Any]:
        return iter(self._instance.items())

    def store(self, filename: Optional[str] = None) -> None:
        """Save environment into memory or file.

        :param filename: a string containing the path of the filename in which
            the environment will be saved. If set to None the environment is
            saved into memory in a stack like structure.
        """
        # Store environment variables
        self.environ = os.environ.copy()

        # Store cwd
        self.cwd = os.getcwd()

        if filename is None:
            self._context.append(pickle.dumps(self._instance))
        else:
            with open(filename, "wb+") as fd:
                pickle.dump(self._instance, fd)

    def restore(self, filename: Optional[str] = None) -> None:
        """Restore environment from memory or a file.

        :param filename: a string containing the path of the filename from
            which the environment will be restored. If set to None the
            environment is pop the last saved one
        """
        if filename is None:
            # We are restoring from memory.  In that case, just double-check
            # that we did store the Env object in memory beforehand (using
            # the store method).
            assert self.environ is not None

        if filename is None and self._context:
            self._instance = pickle.loads(self._context[-1])
            self._context = self._context[:-1]
        elif filename is not None:
            with open(filename, "rb") as fd:
                self._instance = pickle.load(fd)
        else:
            return

        if TYPE_CHECKING:
            assert self.environ is not None

        # Restore environment variables value
        # Do not use os.environ = self.environ.copy()
        # or it will break the os.environ object and child process
        # will get the old environment.
        for k in list(os.environ.keys()):
            if os.environ[k] != self.environ.get(k, None):
                del os.environ[k]
        for k in self.environ:
            if os.environ.get(k, None) != self.environ[k]:
                os.environ[k] = self.environ[k]

        # Restore current directory
        if TYPE_CHECKING:
            assert self.cwd is not None
        os.chdir(self.cwd)



from setuptools import setup, find_packages

import os

install_requires = [
    "colorama",
    "pyyaml",
    "python-dateutil",
    "requests",
    "requests_toolbelt",
    "tqdm",
    "stevedore>1.20.0",
]

extras_require = {"config": ["tomlkit", "typeguard"]}

for p in ("darwin", "linux", "linux2", "win32"):
    platform_string = ":sys_platform=='%s'" % p
    extras_require[platform_string] = ["psutil"]
    if p in ("linux", "linux2"):
        extras_require[platform_string].append("ld")

# Get e3 version from the VERSION file.
version_file = os.path.join(os.path.dirname(__file__), "VERSION")
with open(version_file) as f:
    e3_version = f.read().strip()

with open(os.path.join(os.path.dirname(__file__), "README.md")) as f:
    long_description = f.read()

setup(
    name="e3-core",
    version=e3_version,
    url="https://github.com/AdaCore/e3-core",
    license="GPLv3",
    author="AdaCore",
    author_email="info@adacore.com",
    description="E3 core. Tools and library for building and testing software",
    long_description=long_description,
    long_description_content_type="text/markdown",
    namespace_packages=["e3"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development :: Build Tools",
    ],
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    package_data={"e3": ["py.typed", "os/data/rlimit-*"]},
    install_requires=install_requires,
    extras_require=extras_require,
    entry_points={
        "e3.anod.sandbox.sandbox_action": [
            "exec = e3.anod.sandbox.action:SandBoxExec",
            "create = e3.anod.sandbox.action:SandBoxCreate",
            "show-config = e3.anod.sandbox.action:SandBoxShowConfiguration",
            "migrate = e3.anod.sandbox.migrate:SandBoxMigrate",
        ],
        "e3.event.handler": [
            "smtp = e3.event.handler.smtp:SMTPHandler",
            "logging = e3.event.handler.logging:LoggingHandler",
            "file = e3.event.handler.file:FileHandler",
            "s3 = e3.event.handler.s3:S3Handler",
        ],
        "e3.store": [
            "http-simple-store = e3.store.backends." "http_simple_store:HTTPSimpleStore"
        ],
        "e3.store.cache.backend": [
            "file-cache = e3.store.cache.backends.filecache:FileCache"
        ],
        "sandbox_scripts": ["anod = e3.anod.sandbox.scripts:anod"],
        "console_scripts": [
            "e3 = e3.sys:main",
            "e3-sandbox = e3.anod.sandbox.main:main",
        ],
    },
)



from __future__ import annotations

import sys
from functools import wraps
from typing import TYPE_CHECKING

import e3.log
import e3.store
from e3.anod.spec import AnodError, has_primitive
from e3.env import BaseEnv
from e3.error import E3Error

logger = e3.log.getLogger("e3.anod.driver")

if TYPE_CHECKING:
    from typing import Any, Callable, Literal, TypeVar
    from e3.anod.spec import Anod
    from e3.store.backends.base import Store
    from e3.anod.sandbox import SandBox
    from e3.anod.loader import AnodSpecRepository

    F = TypeVar("F", bound=Callable[..., Any])


def primitive_check() -> Callable[[F], F]:
    def decorator(func: F) -> F:
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            if not has_primitive(self.anod_instance, func.__name__):
                raise AnodError("no primitive %s" % func.__name__)
            elif self.anod_instance.build_space is None:
                raise AnodError(".activate() has not been called")
            return func(self, *args, **kwargs)

        return wrapper  # type: ignore

    return decorator


class AnodDriver:
    def __init__(self, anod_instance: Anod, store: Store):
        """Initialize the Anod driver for a given Anod instance.

        :param anod_instance: an Anod instance
        :param store: the store backend for accessing source and
            binary packages
        """
        self.anod_instance = anod_instance
        self.store = store

    def activate(self, sandbox: SandBox, spec_repository: AnodSpecRepository) -> None:
        self.anod_instance.bind_to_sandbox(sandbox)

        self.anod_instance.log = e3.log.getLogger("spec." + self.anod_instance.uid)

        for e in getattr(self.anod_instance, "%s_deps" % self.anod_instance.kind, ()):
            if isinstance(e, self.anod_instance.Dependency):
                dep_class = spec_repository.load(e.name)
                dep_instance = dep_class(
                    qualifier=e.qualifier,
                    kind=e.kind,
                    env=e.env(self.anod_instance, BaseEnv.from_env()),
                )
                self.anod_instance.deps[e.local_name] = dep_instance
        e3.log.debug("activating spec %s", self.anod_instance.uid)

    def call(self, action: str) -> Any:
        """Call an Anod action.

        :param action: the action (build, install, test, ...)
        """
        return getattr(self, action, self.unknown_action)()

    @staticmethod
    def unknown_action() -> Literal[False]:
        logger.critical("unknown action")
        return False

    @primitive_check()
    def download(self):
        """Run the download primitive."""
        # First check whether there is a download primitive implemented by
        # the Anod spec.
        self.anod_instance.build_space.create(quiet=True)
        download_data = self.anod_instance.download()
        if download_data is None:
            raise AnodError("no download metadata returned by the download primitive")
        try:
            metadata = self.store.get_resource_metadata(download_data)
        except E3Error as err:
            self.anod_instance.log.critical(err)
            raise AnodError(
                "cannot get resource metadata from store", origin=self.anod_instance.uid
            ).with_traceback(sys.exc_info()[2])
        else:
            self.store.download_resource(
                metadata, self.anod_instance.build_space.binary_dir
            )



import numbers
import os
import stat
import sys
import tempfile
import time

import e3.fs
import e3.os.fs

import pytest


def test_cd():
    with pytest.raises(e3.os.fs.OSFSError) as err:
        e3.os.fs.cd("doesnotexist")

    assert "doesnotexist" in str(err)


@pytest.mark.xfail(
    sys.platform == "win32", reason="not completely supported on windows"
)
def test_chmod():
    os.umask(0o022)
    e3.os.fs.touch("a")

    def check_mode(filename, mode):
        fmode = stat.S_IMODE(os.stat(filename).st_mode)
        assert fmode == mode, "{} != {}".format(oct(fmode), oct(mode))

    with pytest.raises(e3.os.fs.OSFSError) as err:
        e3.os.fs.chmod("666", "a")
    assert "not supported" in str(err)

    os.chmod("a", 0o666)
    check_mode("a", 0o666)

    e3.os.fs.chmod("u-w", "a")
    check_mode("a", 0o466)

    e3.os.fs.chmod("a+wx", "a")
    check_mode("a", 0o777)

    e3.os.fs.chmod("-w", "a")
    check_mode("a", 0o577)

    e3.os.fs.chmod("o=u", "a")
    check_mode("a", 0o575)

    e3.os.fs.chmod("u=g", "a")
    check_mode("a", 0o775)

    e3.os.fs.chmod("u+", "a")
    check_mode("a", 0o775)

    e3.os.fs.chmod("g=rw", "a")
    check_mode("a", 0o765)

    e3.os.fs.chmod("a=rwx", "a")
    check_mode("a", 0o777)

    os.chmod("a", 0o666)

    e3.os.fs.chmod("u=rwx,g=rwx", "a")
    check_mode("a", 0o776)

    e3.os.fs.chmod("u=rwx,g=rwx,o=rwx", "a")
    check_mode("a", 0o777)


@pytest.mark.xfail(
    sys.platform == "win32", reason="windows specific rm not yet added to e3-core"
)
def test_rm():
    base = tempfile.mkdtemp()
    try:
        bc = os.path.join(base, "b", "c")
        os.makedirs(bc)
        e3.os.fs.touch(os.path.join(bc, "d"))
        e3.os.fs.chmod("a-w", bc)

        with pytest.raises(OSError):
            e3.os.fs.safe_remove(os.path.join(bc, "d"))

        # Use the high-level rm function to force the deletion
        e3.fs.rm(os.path.join(bc, "d"))

        assert not os.path.exists(os.path.join(bc, "d"))
        e3.os.fs.safe_rmdir(bc)
        assert not os.path.exists(bc)

    finally:
        e3.fs.rm(base, True)


def test_mv():
    os.makedirs("a")
    e3.os.fs.mv("a", "b")
    assert os.path.isdir("b")

    os.makedirs("c")
    e3.os.fs.mv("b", "c")
    assert os.path.isdir(os.path.join("c", "b"))

    e3.os.fs.touch("d")
    os.makedirs("dest")
    e3.os.fs.mv("d", "dest")

    assert os.path.isfile(os.path.join("dest", "d"))


def test_df():
    cwd = os.getcwd()
    statfs = e3.os.fs.df(cwd)
    assert isinstance(statfs, numbers.Integral)
    statfs = e3.os.fs.df(cwd, True)
    assert all(isinstance(elt, numbers.Integral) for elt in statfs)


def test_maxpath():
    maxPath = e3.os.fs.max_path()
    assert isinstance(maxPath, numbers.Integral)


def test_touch():
    e3.os.fs.touch("a")
    assert os.path.exists("a")

    now = time.time()
    os.utime("a", (now - 10000, now))

    e3.os.fs.touch("a")

    assert os.stat("a").st_atime - now < 100


def test_which():
    path_to_e3 = e3.os.fs.which("e3")
    assert os.path.isfile(path_to_e3)

    assert e3.os.fs.which(path_to_e3) == path_to_e3



import e3.os.process
import e3.sys


def test_main():
    e3_tool = e3.sys.python_script("e3")
    assert e3.os.process.Run(e3_tool + ["--version"]).status == 0
    assert "Everything OK!" in e3.os.process.Run(e3_tool + ["--check"]).out



import os

from e3.anod.buildspace import BuildSpace
from e3.fs import mkdir, rm
from e3.os.fs import touch

import pytest


def test_reset():
    """Verify that BuildSpace reset() delete expected content."""
    bs = BuildSpace(root_dir=os.getcwd())
    bs.create()
    for name in bs.dirs:
        touch(os.path.join(bs.subdir(name), "a"))

    # Reset delete all dirs except log and result by default
    bs.reset()
    for name in bs.dirs:
        exists = os.path.exists(os.path.join(bs.subdir(name), "a"))
        if name in ("log", "results"):
            assert exists
        else:
            assert not exists


def test_subdir():
    """Check error handling of subdir."""
    bs = BuildSpace(root_dir=os.getcwd())
    with pytest.raises(ValueError):
        bs.subdir("foo")


def test_reset_tmp_dir():
    """Check that the tmp_dir is reset when the build space is created.

    REQ-EC19.
    """
    bs = BuildSpace(root_dir=os.getcwd())
    marker = os.path.join(bs.subdir(name="tmp"), "deleteme")
    mkdir(bs.tmp_dir)
    touch(marker)
    assert os.path.exists(marker)
    bs.create()
    assert not os.path.exists(marker)


def test_build_space_exists():
    """Test the BuildSpace.exists method."""
    bs_name = os.path.abspath("foo")
    bs = BuildSpace(bs_name)

    # First, verify the behavior when the buildspace directory
    # doesn't even exist.
    assert not os.path.exists(bs_name), bs_name
    assert bs.exists() is False

    # Next, create the directory, but without anything in it.
    # In particular, the marker file isn't present, so
    # is_buildspace should still return False for that directory.
    mkdir(bs_name)
    assert bs.exists() is False

    # Create the buildspace, and then verify that is_buildspace
    # then returns True.
    bs.create()
    assert bs.exists() is True

    # Verify that we also return False if one of the subdirectories
    # is missing. To do that, first verify that the subdirectory
    # we picked does exist, then delete it, before observing
    # whether BuildSpace.exists now return False or not.
    one_subdir = bs.subdir(bs.DIRS[0])
    assert os.path.isdir(one_subdir)
    rm(one_subdir, recursive=True)
    assert bs.exists() is False



"""Convert to Anod newer API."""


from __future__ import annotations

from e3.anod.sandbox import SandBoxError
from e3.anod.sandbox.action import SandBoxAction
from e3.os.fs import cd


class SandBoxMigrate(SandBoxAction):

    name = "migrate"
    help = "Migrate to a newer anod API version"
    require_sandbox = False

    def add_parsers(self) -> None:
        self.parser.add_argument("api_version", help="API version")
        self.parser.add_argument("spec_dir", help="Path to spec directory")

    def run(self, args):
        if args.api_version != "1.5":
            raise SandBoxError("Only 1.5 is supported for now")
        cd(args.spec_dir)
        migrate_v1_5()


def migrate_v1_5() -> None:
    """Migrate to API 1.5.

    Move all .yaml files in subdirs, when the name of a yaml
    file matches a spec name move the yaml file in
    <spec name>/config.yaml

    When there is an additional suffix, move it to

    <spec name>/<suffix>.yaml

    Make sure to run e3-plan-checker after running this script.
    """
    from glob import glob
    import os

    for f in glob("*.yaml"):
        print("looking at %s" % f)
        name, _ = os.path.splitext(os.path.basename(f))
        if os.path.exists(name + ".anod"):
            try:
                os.mkdir(name)
            except OSError:
                pass
            try:
                os.rename(name + ".yaml", os.path.join(name, "config.yaml"))
            except OSError:  # defensive code
                print("error for %s" % name)
        elif "-" in name:
            print("suffix detected in %s" % f)
            prefix, suffix = name.rsplit("-", 1)
            if not os.path.exists(prefix + ".anod"):
                prefix, suffix2 = prefix.rsplit("-", 1)
                suffix = suffix2 + "-" + suffix
            try:
                try:
                    os.mkdir(prefix)
                except OSError:  # defensive code
                    pass
                os.rename(name + ".yaml", os.path.join(prefix, "%s.yaml" % suffix))
            except Exception as er:  # defensive code
                print(f"error for {name}.yaml {prefix} {suffix}")
                print(er)
        else:
            print("unknown yaml file %s.yaml" % name)



from __future__ import annotations

import os
import smtplib
from email.message import Message

import e3.log
import e3.os.process
from e3.error import E3Error

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import List, Optional

logger = e3.log.getLogger("net.smtp")

system_sendmail_fallback = False


def sendmail(
    from_email: str,
    to_emails: List[str],
    mail_as_string: str,
    smtp_servers: List[str],
    max_size: int = 20,
    message_id: Optional[str] = None,
) -> bool:
    """Send an email with stmplib.

    Fallback to /usr/lib/sendmail or /usr/sbin/sendmail if
    ``e3.net.smtp.system_sendmail_fallback`` is set to True (default False).

    :param from_email: the address sending this email (e.g. user@example.com)
    :param to_emails: A list of addresses to send this email to.
    :param mail_as_string: the message to send (with headers)
    :param smtp_servers: list of smtp server names (hostname), in case of
       exception on a server, the next server in the list will be tried
    :param max_size: do not send the email via smptlib if bigger than
        'max_size' Mo.
    :param message_id: the message id (for debugging purposes)

    :return: boolean (sent / not sent)

    We prefer running smtplib so we can manage the email size.
    We run sendmail in case it fails, assuming the max_size on the system
    is high enough - the advantage of sendmail is that it queues the
    email and retries a few times if the target server is unable
    to receive it.
    """
    mail_size = float(len(mail_as_string)) / (1024 * 1024)
    if mail_size >= max_size:
        # Message too big
        logger.error("!!! message file too big (>= %d Mo): %f Mo", max_size, mail_size)
        return False

    def system_sendmail() -> bool:
        """Run the system sendmail."""
        if system_sendmail_fallback:
            for sendmail_bin in (
                "/usr/lib/sendmail",
                "/usr/sbin/sendmail",
            ):  # all: no cover
                if os.path.exists(sendmail_bin):
                    p = e3.os.process.Run(
                        [sendmail_bin] + to_emails,
                        input="|" + mail_as_string,
                        output=None,
                    )
                    return p.status == 0

        # No system sendmail, return False
        return False

    smtp_class = (
        smtplib.SMTP_SSL
        if "smtp_ssl" in os.environ.get("E3_ENABLE_FEATURE", "").split(",")
        else smtplib.SMTP
    )

    for smtp_server in smtp_servers:
        try:
            s = smtp_class(smtp_server)
        except (OSError, smtplib.SMTPException) as e:
            logger.debug(e)
            logger.debug("cannot connect to smtp server %s", smtp_server)
            continue
        else:
            try:
                if not s.sendmail(from_email, to_emails, mail_as_string):
                    # sendmail returns an empty dictionary if the message
                    # was accepted for delivery to all addresses
                    break
                continue
            except (OSError, smtplib.SMTPException) as e:
                logger.debug(e)
                logger.debug("smtp server error: %s", smtp_server)
                continue
            finally:
                try:
                    s.quit()
                    logger.debug("smtp quit")
                except (OSError, smtplib.SMTPException):
                    # The message has already been delivered, ignore all errors
                    # when terminating the session.
                    pass

    else:
        logger.debug("no valid smtp server found")
        if not system_sendmail():
            return False

    if message_id is not None:
        logger.debug("Message-ID: %s sent successfully", message_id)

    return True


def send_message(
    from_email: str,
    to_emails: List[str],
    subject: str,
    content: str,
    smtp_servers: List[str],
) -> None:
    """Send an e-mail message.

    :param from_email: the address sending this email (e.g. user@example.com)
    :param to_emails: A list of addresses to send this email to
    :param subject: the e-mail's subject
    :param content: the e-mail's content
    """
    msg = Message()
    msg["To"] = ", ".join(to_emails)
    msg["From"] = from_email
    msg["Subject"] = subject
    msg.set_payload(content, "utf-8")

    if not sendmail(
        from_email=from_email,
        to_emails=to_emails,
        mail_as_string=msg.as_string(),
        smtp_servers=smtp_servers,
    ):
        raise E3Error(f"error when sending email {subject}")



from __future__ import annotations

import os
import struct
from ctypes import (
    c_wchar_p,
    create_string_buffer,
    create_unicode_buffer,
    pointer,
    sizeof,
)
from ctypes.wintypes import HANDLE
from typing import Any, Callable, Optional

import e3.log
from e3.os.windows.native_api import (
    NT,
    Access,
    FileAttribute,
    FileInfo,
    IOStatusBlock,
    NTException,
    ObjectAttributes,
    OpenOptions,
    Share,
    Status,
    UnicodeString,
)

logger = e3.log.getLogger("os.windows.fs")


class WithOpenFile:
    def __init__(
        self,
        desired_access: Optional[int] = None,
        shared_access: Optional[int] = None,
        open_options: Optional[int] = None,
    ):
        self.desired_access = desired_access
        self.shared_access = shared_access
        self.open_options = open_options

    def __call__(self, f):
        def wrapper(obj, *args, **kwargs):
            should_open = not obj.handle
            if should_open:
                obj.open(self.desired_access, self.shared_access, self.open_options)

            try:
                return f(obj, *args, **kwargs)
            finally:
                if should_open:
                    obj.close()

        return wrapper


class NTFile:
    """Class to manipulate Win32 paths.

    :ivar path: the full pathname
    :ivar nt_filename: UnicodeString object representing either the full path
        or the basename of the file (when inititialized using a parent object)
    :ivar handle: handle to the file. It is set to None if no call to open
        has been done
    :ivar io_status: structure returned by most of low level API call
    :ivar basic_info: ObjectAttributes object associated with the file
    """

    def __init__(self, filename: str, parent: Optional[NTFile] = None):
        """Initialize a NTFile object.

        :param filename: path to the file or basename if parent is not None
        :param parent: the parent NTFile object
        """
        self.handle = None  # type: Optional[HANDLE]
        if parent is None:
            self.path = os.path.abspath(filename)
            self.nt_filename = UnicodeString("\\??\\%s" % self.path)
            self.parent_handle = None
        else:
            assert "\\" not in filename and "/" not in filename
            self.path = os.path.join(parent.path, filename)
            self.nt_filename = UnicodeString(str(filename))
            self.parent_handle = parent.handle

        self.io_status = IOStatusBlock()
        self.basic_info = FileInfo.Basic()

        self.attr = ObjectAttributes(self.nt_filename, parent=self.parent_handle)

        self.desired_access = Access.READ_DATA | Access.READ_ATTRS

        # By default don't block access to other processes
        self.shared_access = Share.ALL

        # By default we set BACKUP_INTENT to be able to bypass ACLs in case
        # the user has sufficient rights. It has no effect otherwise
        self.open_options = OpenOptions.BACKUP_INTENT

    def __str__(self):
        result = [
            "{:<20}: {}".format("path", self.path),
            "{:<20}: {}".format("nt_filename", self.nt_filename),
        ]
        return "\n".join(result)

    def open(
        self,
        desired_access: Optional[int] = None,
        shared_access: Optional[int] = None,
        open_options: Optional[int] = None,
    ) -> None:
        """Open file.

        :param desired_access: desired access
            (see e3.os.windows.native_api.Access)
        :param shared_access: sharing parameters
            (see e3.os.windows.native_api.Shared)
        :param open_options: open options
            (see e3.os.windows.native_api.OpenOptions)
        """
        if desired_access is None:
            desired_access = self.desired_access
        if shared_access is None:
            shared_access = self.shared_access
        if open_options is None:
            open_options = self.open_options

        self.handle = HANDLE()

        open_file: Callable = NT.OpenFile  # type: ignore

        status = open_file(
            pointer(self.handle),
            desired_access,
            pointer(self.attr),
            pointer(self.io_status),
            shared_access,
            open_options,
        )
        if status < 0:
            raise NTException(
                status=status,
                message="cannot open file %s" % self.path,
                origin="NTFile.open",
            )

    def close(self):
        """Close file.

        If a call has been done to open, close the handle associated
        with the file and reset the handle to None.
        """
        if self.handle:
            NT.Close(self.handle)
            self.handle = None

    @property
    def volume_path(self) -> str:
        """Retrieve path to the volume containing the file.

        :return: path to a windows volume after junction resolution
        :raise: NTException
        """
        result = create_unicode_buffer(1024)

        get_vol_name: Callable = NT.GetVolumePathName  # type: ignore
        status = get_vol_name(c_wchar_p(self.path), result, 1024)
        if not status:
            raise NTException(
                status=status,
                message="cannot find volume for %s" % self.path,
                origin="NTFile.volume_path",
            )
        return result.value

    @property  # type: ignore
    @WithOpenFile(Access.READ_ATTRS)
    def uid(self) -> int:
        """Retrieve the ID of the file.

        On NTFS system we are sure that this ID is unique on the given volume

        :return: the uid
        :raise: NTException
        """
        result = FileInfo.Internal()

        query_info_file: Callable = NT.QueryInformationFile  # type: ignore

        status = query_info_file(
            self.handle,
            pointer(self.io_status),
            pointer(result),
            sizeof(result),
            FileInfo.Internal.class_id,
        )
        if status < 0:  # defensive code
            # we should already have raised an error here when trying
            # to open the file
            raise NTException(
                status=status, message="cannot find file uid", origin="NTFile.uid"
            )

        return result.index_number

    def read_attributes_internal(self):
        """Retrieve file basic attributes (internal function).

        The function is used internally to check file basic attributes
        (kind of entry and windows attributes such as readonly). Retrieved
        attributes are stored in basic_info Python attribute. It requires
        less rights than the read_attributes method.
        """
        status = NT.QueryAttributesFile(pointer(self.attr), pointer(self.basic_info))
        if status < 0:
            raise NTException(
                status=status,
                message="cannot query attributes %s" % self.path,
                origin="NTFile.read_attributes_internal",
            )

    @WithOpenFile(Access.READ_ATTRS)
    def read_attributes(self):
        """Retrieve file basic information.

        It updates the basic_info attribute including timestamp information

        :raise: NTException
        """
        result = FileInfo.Basic()
        status = NT.QueryInformationFile(
            self.handle,
            pointer(self.io_status),
            pointer(result),
            sizeof(result),
            FileInfo.Basic.class_id,
        )
        if status < 0:  # defensive code
            # we should already have raised an error here when trying
            # to open the file
            raise NTException(
                status=status,
                message="cannot read attributes",
                origin="NTFile.read_attributes",
            )
        self.basic_info = result
        return result

    @WithOpenFile(Access.WRITE_ATTRS | Access.READ_ATTRS)
    def write_attributes(self):
        """Update file attributes.

        :raise: NTException
        """
        set_infofile: Callable = NT.SetInformationFile
        status = set_infofile(
            self.handle,
            pointer(self.io_status),
            pointer(self.basic_info),
            sizeof(self.basic_info),
            FileInfo.Basic.class_id,
        )
        if status < 0:
            raise NTException(
                status=status,
                message="cannot write attributes to %s" % self.path,
                origin="NTFile.write_attributes",
            )
        self.read_attributes()

    @property
    def is_dir(self) -> bool:
        """Check if file is a directory.

        Note that read_attributes methods should be called at least once
        otherwise the function will return False.

        :return: True if the file is a directory, False otherwise
        """
        return self.basic_info.file_attributes.attr & FileAttribute.DIRECTORY > 0

    @property
    def is_readonly(self) -> bool:
        """Check if file is readonly.

        Note that read_attributes methods should be called at least once
        otherwise the function will return False.

        :return: True if readonly, False otherwise
        """
        return self.basic_info.file_attributes.attr & FileAttribute.READONLY > 0

    @property
    def trash_path(self) -> str:
        """Return path in which the file can move safely for deletion.

        On NTFS filesystem we are sure that the path is unique and thus
        that no existing file can exist at that location.

        :return: a path
        """
        return os.path.join(self.volume_path, "tmp", "Trash", str("%016X" % self.uid))

    @WithOpenFile(Access.DELETE)
    def rename(self, filename: str, replace: bool = False) -> None:
        """Move file.

        :param filename: target location
        :param replace: if True replace the target file if it exists
        :raise: NTException
        """
        file_target = "\\??\\%s" % os.path.abspath(filename)
        target = file_target.encode("utf_16_le")
        s = "?PL%ss" % len(target)
        b = create_string_buffer(struct.calcsize(s))
        b.raw = struct.pack(s, replace, 0, len(target), target)

        set_infofile: Callable = NT.SetInformationFile  # type: ignore

        status = set_infofile(
            self.handle,
            pointer(self.io_status),
            b,
            struct.calcsize(s),
            FileInfo.Rename.class_id,
        )
        if status < 0:
            raise NTException(
                status=status,
                message=f"move of {self.path} to {filename} failed",
                origin="NTFile.rename",
            )

    def move_to_trash(self):
        """Move file to trash location.

        :raise: NTException
        """
        self.rename(self.trash_path, True)

    @WithOpenFile(Access.DELETE, Share.DELETE)
    def dispose(self):
        """Remove the file (low level).

        The remove is effective on call to close method
        """
        fd = FileInfo.Disposition(1)
        status = NT.SetInformationFile(
            self.handle,
            pointer(self.io_status),
            pointer(fd),
            sizeof(FileInfo.Disposition),
            FileInfo.Disposition.class_id,
        )
        if status < 0:
            raise NTException(
                status=status, message="cannot dispose", origin="NTFile.dispose"
            )

    @WithOpenFile(
        Access.LIST_DIRECTORY | Access.SYNCHRONIZE,
        None,
        OpenOptions.SYNCHRONOUS_IO_NON_ALERT,
    )
    def iterate_on_dir(self, fun: Callable, default_result: Any = None) -> Any:
        """Iterate on directory.

        :param fun: function called on each entry (. are .. are skipped)
        :param default_result: default return value
        :return: last return value or fun or default_result
        """
        result = default_result
        s_size = struct.calcsize("LLL")
        b_size = 100 * 1024
        b = create_string_buffer(b_size)

        query_dir_file: Callable = NT.QueryDirectoryFile  # type: ignore

        status = query_dir_file(
            self.handle,
            None,
            None,
            None,
            pointer(self.io_status),
            b,
            b_size,
            FileInfo.Names.class_id,
            False,
            None,
            True,
        )
        if status == Status.NO_MORE_FILES:  # defensive code
            # In theory this case should not occurs at it means that the
            # directory does not even have the . and .. entries. In practice
            # it can occurs (probably because of an intermediate state).
            # In that case behave as if the directory is empty
            return result

        if status < 0:
            raise NTException(
                status=status,
                message="can't read dir %s" % self.path,
                origin="NTFile.iterate_on_dir",
            )

        while status >= 0 and status != Status.NO_MORE_FILES:
            pos = 0
            while True:
                off, _, size = struct.unpack_from("LLL", b.raw, pos)
                name = b.raw[pos + s_size : pos + s_size + size].decode("utf-16-le")
                if name != "." and name != "..":
                    result, should_exit = fun(name, self)
                    if should_exit:
                        return result

                if off == 0:
                    break
                pos += off

            status = query_dir_file(
                self.handle,
                None,
                None,
                None,
                pointer(self.io_status),
                b,
                b_size,
                FileInfo.Names.class_id,
                False,
                None,
                False,
            )
        return result

    @property  # type: ignore
    @WithOpenFile(
        Access.LIST_DIRECTORY | Access.SYNCHRONIZE,
        None,
        OpenOptions.SYNCHRONOUS_IO_NON_ALERT,
    )
    def is_dir_empty(self) -> bool:
        """Check if dir is empty.

        :return: True if the directory is empty
        :raise: NTException
        """

        def check_file(filename, parent):
            f = NTFile(filename, parent)
            try:
                f.read_attributes_internal()
            except NTException:
                return True, False
            return False, True

        return self.iterate_on_dir(check_file, True)

    def unlink(self):
        """Remove file safely.

        :raise: NTException
        """
        open_options = self.open_options
        is_in_trash = False

        # First we need to check that file is not mark readonly
        try:
            self.read_attributes_internal()
        except NTException as e:
            if e.status == Status.OBJECT_NAME_NOT_FOUND:
                return
            elif e.status == Status.DELETE_PENDING:
                return
            else:
                raise

        if self.is_readonly:
            # Try to remove the readonly flag
            self.basic_info.file_attributes.attr &= ~FileAttribute.READONLY
            self.write_attributes()

        # set our access modes
        desired_access = Access.DELETE
        shared_access = Share.DELETE
        if self.is_dir:
            desired_access |= Access.LIST_DIRECTORY | Access.SYNCHRONIZE
            open_options |= OpenOptions.SYNCHRONOUS_IO_NON_ALERT

        try_counter = 10
        # Open the file for deletion
        while try_counter > 0:
            try:
                self.open(desired_access, shared_access, open_options)
                break
            except NTException as e:
                if e.status == Status.SHARING_VIOLATION:
                    # File is already open elsewhere for a non delete operation
                    # Try a few times to open it with relaxed share settings
                    shared_access = Share.ALL
                elif e.status == Status.DELETE_PENDING:  # defensive code
                    # file is already pending deletion (just after our call
                    # to read_attributes) so consider the deletion
                    # is done and return
                    return
                else:  # defensive code
                    # We don't know what to do here so just fail
                    raise

            # Wait a few ms before attempting again to open the file
            NT.Sleep(5)
            try_counter -= 1

        if try_counter == 0:
            raise NTException(
                status=1,
                message="cannot open file %s for deletion" % self.path,
                origin="NTFile.unlink",
            )

        # From there we assume that the file has been opened
        try:
            if shared_access == Share.ALL:
                # The file is also opened elsewhere for a non delete operation
                # In that case we will try to move it to the trash
                # first check that the directory is empty
                if self.is_dir and not self.is_dir_empty:
                    raise NTException(
                        status=1,
                        message="directory not empty: %s" % self.path,
                        origin="NTFile.unlink",
                    )

                self.move_to_trash()
                is_in_trash = True

            # If the file has been moved away then we try to delete it in the
            # trash but it is not necessary to try as hard as the goal is to
            # just keep the trash space used as low as possible.
            if is_in_trash:
                try_counter = 5
            else:
                try_counter = 20

            while try_counter > 0:
                try:
                    self.dispose()
                    break
                except NTException as e:
                    if e.status == Status.DIRECTORY_NOT_EMPTY:
                        # The directory is not empty but that might be because
                        # of remaining files in PENDING_DELETE status. Our
                        # is_dir_empty method return empty if the directory
                        # contains only files pending for deletion
                        if not self.is_dir_empty:
                            raise NTException(
                                status=e.status,
                                message="dir %s is not empty" % self.path,
                                origin="NTFile.unlink",
                            )
                    elif e.status == Status.CANNOT_DELETE:  # defensive code
                        # At this stage we are sure that the file is not
                        # read_only but it seems that we can get this error
                        # when the file has been mapped to memory.
                        if not is_in_trash:
                            try:
                                self.move_to_trash()
                                is_in_trash = True
                                try_counter = min(try_counter, 5)
                            except NTException:
                                pass
                    else:  # defensive code
                        # Unknown error. If the file has been moved away
                        # consider it success. Otherwise reraise exception
                        if is_in_trash:
                            break
                        raise

                NT.Sleep(5)
                try_counter -= 1

        finally:
            self.close()



from __future__ import annotations

import hashlib
import os

import e3.error


class HashError(e3.error.E3Error):
    pass


def __compute_hash(path: str, kind: str) -> str:
    if not os.path.isfile(path):
        raise HashError(kind, "cannot find %s" % path)

    with open(path, "rb") as f:
        result = getattr(hashlib, kind)()
        while True:
            data = f.read(1024 * 1024)
            if not data:
                break
            result.update(data)
    return result.hexdigest()


def md5(path: str) -> str:
    """Compute md5 hexadecimal digest of a file.

    :param path: path to a file

    :return: the hash of the file content
    :raise HashError: in case of error
    """
    return __compute_hash(path, "md5")


def sha1(path: str) -> str:
    """Compute sha1 hexadecimal digest of a file.

    :param str path: path to a file

    :return: the hash of the file content
    :raise HashError: in case of error
    """
    return __compute_hash(path, "sha1")


def sha256(path: str) -> str:
    """Compute sha256 hexadecimal digest of a file.

    :param str path: path to a file

    :return: the hash of the file content
    :raise HashError: in case of error
    """
    return __compute_hash(path, "sha256")



from __future__ import annotations

import json
import mimetypes
from email import encoders
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.utils import formatdate, make_msgid
from typing import TYPE_CHECKING

import e3.net.smtp
from e3.event import EventHandler

if TYPE_CHECKING:
    from typing import Dict, List, Union
    from e3.event import Event


class SMTPHandler(EventHandler):
    def __init__(
        self,
        subject: str,
        from_addr: str,
        to_addr: str,
        smtp_servers: Union[str, List[str]],
    ):
        """Initialize a SMTP event manager."""
        self.default_subject = subject
        self.from_addr = from_addr
        self.to_addr = to_addr
        if isinstance(smtp_servers, str):
            self.smtp_servers = [smtp_servers]
        else:
            self.smtp_servers = smtp_servers

    @classmethod
    def decode_config(cls, config_str: str) -> Dict[str, Union[str, List[str]]]:
        subject, from_addr, to_addr, smtp_servers = config_str.split(",", 3)
        return {
            "subject": subject,
            "from_addr": from_addr,
            "to_addr": to_addr,
            "smtp_servers": smtp_servers.split(","),
        }

    def encode_config(self) -> str:
        return "{},{},{},{}".format(
            self.subject, self.from_addr, self.to_addr, ",".join(self.smtp_servers),
        )

    @property
    def subject(self) -> str:
        return self.default_subject

    def send_event(self, event: Event) -> bool:
        """Send an event.

        :param event: an event
        :return: True if the event was sent successfully
        """
        json_content = json.dumps(event.as_dict())
        attachments = event.get_attachments()

        mail = MIMEMultipart()
        mail["Subject"] = self.subject
        mail["From"] = self.from_addr
        mail["To"] = self.to_addr
        mail["Date"] = formatdate(localtime=True)
        mail["Message-ID"] = make_msgid()
        mail.preamble = "You will not see this in a MIME-aware mail reader.\n"
        event_json = MIMEBase("application", "json")
        event_json.set_payload(json_content)
        encoders.encode_base64(event_json)
        event_json.add_header("Content-Disposition", "attachment", filename="event")
        mail.attach(event_json)

        result: Dict[str, Dict[str, dict]] = {"attachments": {}}

        for name, (filename, _) in list(attachments.items()):
            ctype, encoding = mimetypes.guess_type(filename)

            if encoding == "gzip" and ctype == "application/x-tar":
                attachment = MIMEBase("application", "x-tar-gz")
            elif encoding is None and ctype is not None:
                attachment = MIMEBase(*ctype.split("/", 1))
            else:
                attachment = MIMEBase("application", "octet-stream")
            with open(filename, "rb") as data_f:
                attachment.set_payload(data_f.read())

            result["attachments"][name] = {
                "path": filename,
                "encoding": encoding,
                "ctype": ctype,
            }

            encoders.encode_base64(attachment)
            attachment.add_header("Content-Disposition", "attachment", filename=name)
            mail.attach(attachment)

        return e3.net.smtp.sendmail(
            self.from_addr,
            mail["To"].split(","),
            mail.as_string(),
            self.smtp_servers,
            message_id=mail["Message-ID"],
        )



import os
import sys

import e3.diff
import e3.fs
import e3.hash
import e3.os.fs

import pytest


def test_cp():
    current_dir = os.getcwd()
    hash_test = os.path.join(current_dir, "hash_test")
    e3.fs.cp(__file__, hash_test)
    assert e3.hash.sha1(__file__) == e3.hash.sha1(hash_test)

    a = os.path.join(current_dir, "a")
    a1 = os.path.join(a, "a1")
    b1 = os.path.join(a, "b", "b1")

    e3.fs.mkdir(a)
    e3.fs.echo_to_file(a1, "a1")
    e3.fs.mkdir(os.path.join(a, "b"))
    e3.fs.echo_to_file(b1, "b1")

    dest = os.path.join(current_dir, "dest")
    e3.fs.mkdir(dest)
    e3.fs.cp(a, dest, recursive=True)
    assert os.path.exists(os.path.join(dest, "a", "a1"))
    assert os.path.exists(os.path.join(dest, "a", "b", "b1"))

    dest2 = os.path.join(current_dir, "dest2")

    with pytest.raises(e3.fs.FSError) as err:
        e3.fs.cp("*.non_existing", dest2)
    assert "can't find files matching" in str(err.value)

    with pytest.raises(e3.fs.FSError) as err:
        e3.fs.cp([a1, b1], dest2)
    assert "target should be a directory" in str(err.value)

    e3.fs.mkdir(dest2)
    e3.fs.cp([a1, b1], dest2)
    assert os.path.exists(os.path.join(dest2, "a1"))
    assert os.path.exists(os.path.join(dest2, "b1"))

    dest3 = os.path.join(current_dir, "dest3")
    e3.fs.mkdir(dest3)
    e3.fs.cp(a, dest3, copy_attrs=False, recursive=True)
    e3.fs.cp(a1, dest3, copy_attrs=False)

    assert os.path.exists(os.path.join(dest3, "a", "a1"))
    assert os.path.exists(os.path.join(dest3, "a", "b", "b1"))
    assert os.path.exists(os.path.join(dest3, "a1"))

    with pytest.raises(e3.fs.FSError):
        e3.fs.cp(a, os.path.join("does", "not", "exist"))


@pytest.mark.skipif(sys.platform == "win32", reason="test using symlink")
def test_cp_symplink():
    e3.os.fs.touch("c")
    os.symlink("c", "c_sym")
    e3.fs.cp("c_sym", "d", preserve_symlinks=True)
    assert os.path.islink("d")


def test_echo():
    dest_file = "echo_test"
    e3.fs.echo_to_file(dest_file, "foo")
    e3.fs.echo_to_file(dest_file, "foo")
    e3.fs.echo_to_file(dest_file, "bar", append=True)
    with open(dest_file) as foobar_f:
        assert foobar_f.read() == "foobar"

    e3.fs.echo_to_file(dest_file, ["line1", "line2"])
    with open(dest_file) as fd:
        assert fd.read().strip() == "line1\nline2"


def test_find():
    d = os.path.dirname(__file__)
    parent_d = os.path.dirname(d)

    result = {os.path.abspath(f) for f in e3.fs.find(parent_d)}
    assert os.path.abspath(__file__) in result
    assert os.path.abspath(d) not in result

    result = e3.fs.find(parent_d, include_dirs=True, include_files=False)
    result = {os.path.abspath(f) for f in result}
    assert os.path.abspath(d) in result
    assert os.path.abspath(__file__) not in result

    result = e3.fs.find(parent_d, include_dirs=True, include_files=True)
    result = {os.path.abspath(f) for f in result}
    assert os.path.abspath(d) in result
    assert os.path.abspath(__file__) in result


def test_ls(caplog):
    e3.os.fs.touch("a")
    e3.fs.ls("a", emit_log_record=True)
    assert "ls a" in caplog.text

    e3.os.fs.touch("b")
    e3.os.fs.touch("c")

    assert e3.fs.ls("*") == ["a", "b", "c"]

    # Reproduce issue #213: add test with generator
    assert e3.fs.ls(k for k in ("a", "c")) == ["a", "c"]


def test_mkdir(caplog):
    e3.fs.mkdir("subdir")
    for record in caplog.records:
        assert "mkdir" in record.msg


def test_mkdir_exists(caplog):
    os.makedirs("subdir")
    e3.fs.mkdir("subdir")
    for record in caplog.records:
        assert "mkdir" not in record.msg


def test_mv():
    for fname in ("a1", "a2", "a3", "1", "2", "3", "11", "12", "13"):
        e3.os.fs.touch(fname)

    e3.fs.mkdir("b")

    e3.fs.mv("a*", "b")
    for fname in ("a1", "a2", "a3"):
        assert os.path.isfile(os.path.join("b", fname))

    e3.fs.mv("1", "b")
    assert os.path.isfile(os.path.join("b", "1"))

    with pytest.raises(e3.fs.FSError):
        e3.fs.mv(("1*", " 2", "3"), "c")

    e3.fs.mkdir("c")
    e3.fs.mv(("1*", "2", "3"), "c")
    for fname in ("2", "3", "11", "12", "13"):
        assert os.path.isfile(os.path.join("c", fname))

    with pytest.raises(e3.fs.FSError):
        e3.fs.mv("d*", "b")


def test_tree_state():
    import time

    current_dir = os.getcwd()
    d = os.path.dirname(os.path.dirname(__file__))
    state = e3.fs.get_filetree_state(d)
    assert isinstance(state, str)

    e3.fs.sync_tree(d, os.path.join(current_dir))
    state2 = e3.fs.get_filetree_state(current_dir)
    assert state != state2

    state3 = e3.fs.get_filetree_state(current_dir)
    assert state2 == state3

    # To ensure that file system resolution is not hidding
    # changes
    time.sleep(2)

    e3.os.fs.touch("toto")
    state4 = e3.fs.get_filetree_state(current_dir)
    assert state4 != state3
    hidden = os.path.join(current_dir, ".h")
    e3.fs.mkdir(hidden)
    state5 = e3.fs.get_filetree_state(current_dir)
    assert state5 == state4
    e3.os.fs.touch(".toto")
    state6 = e3.fs.get_filetree_state(current_dir)
    assert state6 == state5
    state6 = e3.fs.get_filetree_state("toto")
    assert isinstance(state6, str)

    # check that get_filetree_state accept unicode
    state7 = e3.fs.get_filetree_state("toto")
    assert isinstance(state7, str)


@pytest.mark.skipif(sys.platform == "win32", reason="test using symlink")
def test_sync_tree_with_symlinks():
    current_dir = os.getcwd()
    a = os.path.join(current_dir, "a")
    b = os.path.join(current_dir, "b")
    m1 = os.path.join(current_dir, "m1")
    m2 = os.path.join(current_dir, "m2")
    m3 = os.path.join(current_dir, "m3")

    e3.fs.mkdir(m1)
    e3.fs.mkdir(m2)
    e3.fs.mkdir(m3)

    with open(a, "w") as f:
        f.write("a")

    with open(b, "w") as f:
        f.write("b")

    e3.fs.cp(a, os.path.join(m1, "c"))
    os.symlink(b, os.path.join(m2, "c"))
    os.symlink(m2, os.path.join(m3, "c"))

    # we start with m2/c -> b
    # so m2/c and b points to the same content
    assert e3.diff.diff(b, os.path.join(m2, "c")) == ""
    assert e3.diff.diff(b, os.path.join(m1, "c"))
    e3.fs.sync_tree(m1, m2)

    # after the sync tree m1/c = m2/c
    assert e3.diff.diff(os.path.join(m1, "c"), os.path.join(m2, "c")) == ""

    # and m2/c is not a symlink anymore so does not
    # have the same content as b
    assert e3.diff.diff(b, os.path.join(m2, "c"))

    # we start with m3/c -> m2
    assert os.path.exists(os.path.join(m3, "c", "c"))
    e3.fs.sync_tree(m1, m3)
    # after the sync tree m1/c = m3/c
    assert e3.diff.diff(os.path.join(m1, "c"), os.path.join(m3, "c")) == ""

    # and m3/c is not a link to m2
    assert not os.path.exists(os.path.join(m3, "c", "c"))
    assert os.path.exists(os.path.join(m2, "c"))


@pytest.mark.skipif(sys.platform != "win32", reason="test relevant only on win32")
def test_sync_tree_case_insensitive():
    e3.fs.mkdir("test/a")
    e3.fs.mkdir("test/b")
    e3.os.fs.touch("test/a/initial.txt")
    e3.os.fs.touch("test/b/Initial.txt")
    e3.fs.mkdir("test/a/Subdir")
    e3.fs.mkdir("test/b/subdir")
    e3.fs.echo_to_file("test/b/OLD.txt", "Version_Old")
    e3.fs.echo_to_file("test/a/old.txt", "Version_New")
    e3.fs.echo_to_file("test/b/old2.txt", "Version_Old")
    e3.fs.echo_to_file("test/a/OLD2.txt", "Version_New")

    e3.fs.sync_tree("test/a", "test/b")
    assert e3.fs.directory_content("test/b") == e3.fs.directory_content("test/a")

    # Adjust some casing of a file that is up-to-date. Sync_tree should be case
    # preserving and thus adjust the casing
    os.rename("test/b/OLD2.txt", "test/b/Old2.txt")
    e3.fs.sync_tree("test/a", "test/b")
    assert e3.fs.directory_content("test/b") == e3.fs.directory_content("test/a")


def test_sync_tree_preserve_timestamps():
    """Run sync_tree without preserving timestamps."""
    e3.fs.mkdir("a")
    e3.fs.mkdir("b")
    with open("a/content", "w") as f:
        f.write("content")
    with open("a/content2", "w") as f:
        f.write("content2")
    with open("b/content", "w") as f:
        f.write("content")
    with open("b/content2", "w") as f:
        f.write("content1")
    e3.fs.sync_tree("a", "b", preserve_timestamps=False)

    with open("b/content2") as f:
        assert f.read() == "content2"


@pytest.mark.skipif(sys.platform == "win32", reason="test using symlink")
def test_sync_tree_links():
    """Check handling of symbolink links in sync_tree."""
    e3.fs.mkdir("a")
    e3.fs.mkdir("b")
    e3.fs.mkdir("c")
    with open("a/content", "w") as f:
        f.write("content")
    os.symlink(os.path.join(os.getcwd(), "a", "content"), "a/link")
    e3.fs.sync_tree("a", "b", preserve_timestamps=False)

    with open("b/link") as f:
        assert f.read() == "content"


def test_sync_tree_does_not_exist():
    """Check error message when sync_tree source does not exist."""
    with pytest.raises(e3.fs.FSError) as err:
        e3.fs.sync_tree("doesnotexist", "dest")
    assert "doesnotexist does not exist" in str(err)


def test_rm_on_error():
    e3.fs.mkdir("a")
    e3.fs.mkdir("a/b")
    e3.os.fs.touch("a/b/c")
    os.chmod("a/b/c", 0o000)
    os.chmod("a/b", 0o000)

    e3.fs.mkdir("a/d")
    e3.os.fs.touch("a/d/e")
    os.chmod("a/d/e", 0o000)
    os.chmod("a/d", 0o500)

    e3.fs.mkdir("a/f")
    e3.fs.mkdir("a/f/g")
    os.chmod("a/f", 0o500)

    e3.fs.rm("a", True)


def test_rm_list():
    """Call rm with list of files to remove."""
    e3.os.fs.touch("a")
    e3.os.fs.touch("b")
    e3.fs.rm(["a", "b"], glob=False)
    assert not os.path.exists("a")
    assert not os.path.exists("b")


@pytest.mark.skipif(sys.platform == "win32", reason="test using symlink")
def test_rm_symlink_to_dir():
    e3.fs.mkdir("a")
    os.symlink("a", "b")
    e3.fs.rm("b", recursive=True)
    assert not os.path.exists("b")


def test_safe_copy():
    """sync_tree should replace directory by files and fix permissions."""
    # Check that a directory in the target dir is replaced by a file when
    # needed.
    e3.fs.mkdir("a")
    with open("a/f", "w") as f:
        f.write("file")
    e3.fs.mkdir("b")
    e3.fs.mkdir("b/f")
    e3.fs.sync_tree("a", "b")
    assert os.path.isfile("b/f")

    # Check that target file permission are changed to allow copying new
    # content.
    e3.fs.mkdir("c")
    e3.os.fs.touch("c/f")
    os.chmod("c/f", 0o000)
    e3.fs.sync_tree("a", "c")
    with open("c/f") as f:
        assert f.read() == "file"


@pytest.mark.skipif(sys.platform == "win32", reason="test using symlink")
def test_safe_copy_links():
    """sync_tree should replace directory by symlinks when needed."""
    e3.fs.mkdir("a")
    e3.fs.mkdir("a/d")
    os.symlink("f", "a/l")
    e3.fs.mkdir("b")
    e3.fs.mkdir("b/l")
    e3.fs.sync_tree("a", "b")
    assert os.path.islink("b/l")


def test_sync_tree_dir_vs_file():
    """sync_tree should replace file by directory when needed."""
    e3.fs.mkdir("a")
    e3.fs.mkdir("a/d")
    e3.fs.mkdir("b")
    e3.os.fs.touch("b/d")
    e3.fs.sync_tree("a", "b")
    assert os.path.isdir("b/d")


def test_safe_mkdir():
    """sync_tree should copy dir even when no permission in target dir."""
    e3.fs.mkdir("a")
    e3.fs.mkdir("a/a")
    e3.fs.mkdir("b")
    os.chmod("b", 0o000)
    e3.fs.sync_tree("a", "b")
    assert os.path.isdir("b/a")


def test_splitall():
    assert e3.fs.splitall("a/b") == ("a", "b")
    assert e3.fs.splitall("/a") == ("/", "a")
    assert e3.fs.splitall("/a/b") == ("/", "a", "b")
    assert e3.fs.splitall("/a/b/") == ("/", "a", "b")


def test_sync_tree_with_file_list():
    """Test sync_tree with file_list."""
    e3.fs.mkdir("a")
    for x in range(0, 10):
        e3.os.fs.touch("a/%d" % x)

    e3.fs.mkdir("b")
    e3.fs.sync_tree("a", "b", file_list=["3", "7"])
    assert os.path.exists("b/3")
    assert os.path.exists("b/7")
    assert not os.path.exists("b/5")


def test_sync_tree_with_ignore():
    """Test sync_tree with ignore."""
    e3.fs.mkdir("a")
    for x in range(0, 10):
        e3.os.fs.touch("a/%d" % x)

    e3.fs.mkdir("b")
    e3.fs.sync_tree("a", "b", ignore=["/3", "/7"])
    assert not os.path.exists("b/3")
    assert not os.path.exists("b/7")
    assert os.path.exists("b/5")

    e3.os.fs.touch("b/7")
    e3.fs.sync_tree("a", "b", ignore=["/3", "/7"])
    assert os.path.exists("b/7")
    e3.fs.sync_tree("a", "b", ignore=["/3", "/7"], delete_ignore=True)
    assert not os.path.exists("b/7")


def test_extension():
    assert e3.fs.extension("/home/file1.2.txt") == ".txt"
    assert e3.fs.extension("file2.tar.gz") == ".tar.gz"
    assert e3.fs.extension("file2.tar") == ".tar"
    assert e3.fs.extension("file2.tar.bz2") == ".tar.bz2"


def test_directory_content():
    """Test e3.fs.directory_content."""
    e3.fs.mkdir("test1")
    e3.fs.mkdir("test1/test2")
    e3.os.fs.touch(os.path.join("test1", "test1.txt"))
    e3.os.fs.touch(os.path.join("test1", "test2.txt"))
    assert e3.fs.directory_content("test1") == [
        "test1.txt",
        "test2.txt",
        "test2" + os.sep,
    ]
    assert e3.fs.directory_content("test1", unixpath=True) == [
        "test1.txt",
        "test2.txt",
        "test2/",
    ]
    assert e3.fs.directory_content("test1", include_root_dir=True, unixpath=True) == [
        "test1/test1.txt",
        "test1/test2.txt",
        "test1/test2/",
    ]



"""Return the current timezone offset."""


from __future__ import annotations

import sys
from datetime import datetime

from dateutil.tz import gettz


def timezone() -> float:
    """Return current timezone offset in hours.

    :return: offset from utc in hours
    """
    if sys.platform == "win32":  # unix: no cover
        from ctypes import windll, Structure, pointer
        from ctypes.wintypes import DWORD, WCHAR, LONG

        class TIME_ZONE_INFORMATION(Structure):
            _fields_ = [
                ("Bias", LONG),
                ("StandardName", WCHAR * 32),
                ("StandardDate", DWORD * 8),
                ("StandardBias", LONG),
                ("DaylightName", WCHAR * 32),
                ("DaylightDate", DWORD * 8),
                ("DaylightBias", LONG),
            ]

        win_tz = TIME_ZONE_INFORMATION()
        win_tz_pt = pointer(win_tz)
        windll.kernel32.GetTimeZoneInformation(win_tz_pt)
        result = win_tz.Bias
        result = -int(result) // 60
    else:
        # Note that in case timezone cannot be computed then
        # utcoffset can return None.
        offset = datetime.now(gettz()).utcoffset()
        if offset is None:
            result = 0.0  # defensive code
        else:
            result = offset.total_seconds() // 3600

    return float(result)



from e3.collection.toggleable_bool import ToggleableBooleanGroup


def test_toggleable_bools():
    g = ToggleableBooleanGroup()
    for i, v in enumerate((True, True, False, True, False, False)):
        g.add(name="seed{}".format(i), value=v)

    result = []
    result.append([bool(c) for c in g])
    for series in g.shuffle():
        result.append([bool(c) for c in series])
        assert series == list(g)
        assert len(g) == 6

    assert result == [
        [True, True, False, True, False, False],
        [True, True, True, True, True, True],
        [True, True, True, True, True, False],
        [True, True, True, True, False, True],
        [True, True, True, True, False, False],
        [True, True, True, False, True, True],
        [True, True, True, False, True, False],
        [True, True, True, False, False, True],
        [True, True, True, False, False, False],
        [True, True, False, True, True, True],
        [True, True, False, True, True, False],
        [True, True, False, True, False, True],
        [True, True, False, False, True, True],
        [True, True, False, False, True, False],
        [True, True, False, False, False, True],
        [True, True, False, False, False, False],
        [True, False, True, True, True, True],
        [True, False, True, True, True, False],
        [True, False, True, True, False, True],
        [True, False, True, True, False, False],
        [True, False, True, False, True, True],
        [True, False, True, False, True, False],
        [True, False, True, False, False, True],
        [True, False, True, False, False, False],
        [True, False, False, True, True, True],
        [True, False, False, True, True, False],
        [True, False, False, True, False, True],
        [True, False, False, True, False, False],
        [True, False, False, False, True, True],
        [True, False, False, False, True, False],
        [True, False, False, False, False, True],
        [True, False, False, False, False, False],
        [False, True, True, True, True, True],
        [False, True, True, True, True, False],
        [False, True, True, True, False, True],
        [False, True, True, True, False, False],
        [False, True, True, False, True, True],
        [False, True, True, False, True, False],
        [False, True, True, False, False, True],
        [False, True, True, False, False, False],
        [False, True, False, True, True, True],
        [False, True, False, True, True, False],
        [False, True, False, True, False, True],
        [False, True, False, True, False, False],
        [False, True, False, False, True, True],
        [False, True, False, False, True, False],
        [False, True, False, False, False, True],
        [False, True, False, False, False, False],
        [False, False, True, True, True, True],
        [False, False, True, True, True, False],
        [False, False, True, True, False, True],
        [False, False, True, True, False, False],
        [False, False, True, False, True, True],
        [False, False, True, False, True, False],
        [False, False, True, False, False, True],
        [False, False, True, False, False, False],
        [False, False, False, True, True, True],
        [False, False, False, True, True, False],
        [False, False, False, True, False, True],
        [False, False, False, True, False, False],
        [False, False, False, False, True, True],
        [False, False, False, False, True, False],
        [False, False, False, False, False, True],
        [False, False, False, False, False, False],
    ]



import os

from e3.store.backends.http_simple_store import HTTPSimpleStore
from e3.store.cache.backends.filecache import FileCache

import httpretty
import httpretty.core


def test_simple_store(caplog):
    """Basic test of HTTPSimpleStore."""
    with httpretty.core.httprettized():
        httpretty.HTTPretty.allow_net_connect = False

        httpretty.register_uri(
            httpretty.GET,
            "http://test.example",
            body="a body content",
            content_disposition='attachment; filename="foo.tar.gz"',
        )
        query = {
            "url": "http://test.example",
            "sha": "da39a3ee5e6b4b0d3255bfef95601890afd80709",
        }

        store = HTTPSimpleStore({}, None)
        metadata = store.get_resource_metadata(query)
        assert {metadata.url, metadata.sha} == set(query.values())

        current_dir = os.getcwd()
        path = store.download_resource_content(metadata, current_dir)
        assert path is None
        assert "expecting da39a3ee5e6b4b0d3255bfef95601890afd80709 got" in caplog.text

        metadata.sha = "0c8ef1a401f4564abba7b85676464ac4bbb5cb05"
        path = store.download_resource_content(metadata, current_dir)
        assert path is not None


def test_store_with_cache():
    """Test HTTPSimpleStore with a FileCache."""
    fc = FileCache({"cache_dir": os.path.join(os.getcwd(), "cache")})
    with httpretty.core.httprettized():
        httpretty.HTTPretty.allow_net_connect = False

        httpretty.register_uri(
            httpretty.GET,
            "http://test.example",
            body="a body content",
            content_disposition='attachment; filename="foo.tar.gz"',
        )

        query = {
            "url": "http://test.example",
            "sha": "0c8ef1a401f4564abba7b85676464ac4bbb5cb05",
        }

        store = HTTPSimpleStore({}, fc)
        metadata = store.get_resource_metadata(query)
        current_dir = os.getcwd()
        path = store.download_resource(metadata, current_dir)
        assert path is not None
        with open(path) as f:
            assert "a body content" in f.read()

        # Calling twice should return the same result (from cache)
        path2 = store.download_resource(metadata, current_dir)
        assert path == path2

        # invalidate the cache
        cached_data = store.cache_backend.get(metadata.uid)
        with open(cached_data.local_path, "a") as f:
            f.write("-invalid")
        # the cache entry will be deleted and we should have the right result
        path3 = store.download_resource(metadata, current_dir)
        assert path == path3

        # replace the cached_data entry by 0 to raise a TypeError
        store.cache_backend.set(metadata.uid, 0)
        path4 = store.download_resource(metadata, current_dir)
        assert path == path4

        # Verify that the function returns None when the resource
        # does not exist
        metadata.sha = "fff"
        assert store.download_resource(metadata, current_dir) is None



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Callable, Dict, Optional


class EntryPoint:
    """Plan Entry point.

    The class represent a valid entry point in an electrolyt plan. It's used
    also to annotate a given entry point with some additional metadata.
    """

    def __init__(
        self,
        db: Dict[str, EntryPoint],
        fun: Callable[[], None],
        kind: str,
        name: Optional[str] = None,
        description: Optional[str] = None,
    ):
        """Initialize an entry point.

        :param db: dictionary that tracks the list of entry points
        :param fun: function that implement the entry point
        :param kind: name used in the plan for the entry point decorator
        :param name: name of the entry point. If None then the function
            name is used
        :param description: a description of the entry point
        """
        if name is None:
            self.name = fun.__name__
        else:
            self.name = name
        self.fun = fun
        self.kind = kind
        self.description = description
        self.executed = False
        assert self.name not in db, "duplicate entry point %s" % self.name
        db[self.name] = self

    def execute(self) -> None:
        """Execute an entry point."""
        self.executed = True
        self.fun()


class Machine(EntryPoint):
    """Machine entry point.

    The class is used to declare an entry point which will be the default
    function executed on a given server. It also add some information such
    as the platform and the OS version in order to be able to simulate a plan
    execution.
    """

    def __init__(
        self,
        db: Dict[str, EntryPoint],
        fun: Callable[[], None],
        kind: str,
        platform: str,
        version: str,
        site: Optional[str] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
    ):
        """Initialize a Machine entry point.

        :param db: see EntryPoint
        :param fun: see EntryPoint
        :param platform: platform name
        :param version: OS version
        :param site: site indication
        :param name: see EntryPoint
        :param description: see EntryPoint
        """
        super().__init__(db, fun, kind, name, description)
        self.platform = platform
        self.version = version
        self.site = site


def entry_point(
    db: Dict[str, EntryPoint],
    cls: Callable[..., EntryPoint],
    kind: str,
    *args: Any,
    **kwargs: Any
) -> Callable:
    """Entry point decorator.

    Declare an electrolyt entry point (e.g. a machine name, a mailserver
    preset).

    This returns the function with the following new attributes: kind, name,
    description, is_entry_point=True, and executed. Executed is a boolean set
    to False by default and changed to True when the entry point has been
    executed.

    :param db: dictionary where to register the entry point
    :param cls: class of entry point
    :param kind: entry point kind (machine name, mailserver preset, ...)
    :param args: additional information to store with the entry point
    :param kwargs: additional information to store with the entry point
    """

    def entry_point_dec(f, ldb=db, lcls=cls, lkind=kind, largs=args, lkwargs=kwargs):
        lcls(ldb, f, lkind, *largs, **lkwargs)
        return f

    return entry_point_dec



from __future__ import annotations

import re
from typing import Dict


def format_with_dict(pattern: str, values: Dict[str, str]) -> str:
    """Safely format a python string using % and a dictionary for values.

    This method is safer than using directly percent as it will escape
    automatically % in the pattern that cannot be replaced.

    :param pattern: a string that should be formatted
    :param values: a dictionary containing the values of the keys that can be
        replaced
    """
    key_regexp = r"|".join([r"\(%s\)" % k for k in values])
    result = re.sub(r"%%(?!%s)" % key_regexp, r"%%", pattern)
    return result % values


def bytes_as_str(content: bytes) -> str:
    """Safely convert bytes to a python string.

    This function attempts the conversion assuming a utf-8 enconding.
    If this triggers a conversion error, it then falls back to a safe
    representation, where some characters might be escaped.

    :param content: content to be transformed
    :return: a string
    """
    try:
        result = content.decode("utf-8")
    except UnicodeDecodeError:
        result = str(content)[2:-1].replace("\\n", "\n")
    return result



"""Add set of conditional booleans that can be used to explore many combinations.

ToggleableBoolean are meant to be grouped and their value can be toggled to
explore all possible combinations. This can be useful when there is an expression
depending on external values that you want to evaluate without knowing in advance
the external environment.
"""
from __future__ import annotations

from itertools import product
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Iterator, List


class ToggleableBooleanGroup:
    """Contain group of toggleable boolean.

    Add a new boolean by calling .add(<name>, <bool value>). Then generate
    all possible combinations by running .shuffle()::

        group = ToggleableBooleanGroup()
        group.add('first', True)
        group.add('second', False)

        for series in group.shuffle():
            print([str(b) for b in series])

    Will display::

        ['first: True', 'second: True']
        ['first: False', 'second: True']
        ['first: False', 'second: False']
    """

    def __init__(self) -> None:
        self.series: List[ToggleableBoolean] = []

    def __getitem__(self, key: int) -> ToggleableBoolean:
        return self.series[key]

    def __len__(self):
        return len(self.series)

    def shuffle(self) -> Iterator[List[ToggleableBoolean]]:
        """Generate all other possible set of values for all conditional booleans.

        :return: yield a new list of ToggleableBoolean with a different set of value.
            Calling this function until StopIteration is raised will generate all
            possible values except the initial set of value.
        """
        initial_values = tuple(bool(c) for c in self.series)
        for shuffeld_values in product([True, False], repeat=len(initial_values)):
            if shuffeld_values != initial_values:
                # Replace conditional values by these new shuffled values
                for idx, value in enumerate(shuffeld_values):
                    self.series[idx].value = value
                yield self.series

    def add(self, name: str, value: bool) -> ToggleableBoolean:
        """Create a new boolean value and add it to this group.

        :param name: boolean name for debug info
        :param value: boolean value
        """
        result = ToggleableBoolean(name=name, value=value)
        self.series.append(result)
        return result


class ToggleableBoolean:
    """Contain a boolean value that can be toggle in group.

    See ToggleableBooleanGroup.
    """

    def __init__(self, name: str, value: bool) -> None:
        """Set boolean value.

        :param name: boolean name for debug info
        :param value: boolean value
        """
        self.value = value
        self.name = name

    def __bool__(self) -> bool:
        return self.value

    def __str__(self):
        return f"{self.name}: {self.value}"
